###############################################################################
# Name: Cody Precord                                                          #
# Purpose: File abstraction layer and text utilities                          #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Text/Unicode handling functions and File wrapper class

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: ed_txt.py 72103 2012-07-15 15:02:42Z CJP $"
__revision__ = "$Revision: 72103 $"

#--------------------------------------------------------------------------#
# Imports
import sys
import re
import time
import wx
import codecs
import encodings as enclib
import locale
from StringIO import StringIO

# Local Imports
from util import Log
from profiler import Profile_Get
import ed_msg
import ebmlib
import ed_thread

#--------------------------------------------------------------------------#
# Globals

# The default fallback encoding
DEFAULT_ENCODING = locale.getpreferredencoding()
try:
    codecs.lookup(DEFAULT_ENCODING)
except (LookupError, TypeError):
    DEFAULT_ENCODING = 'utf-8'

# File Helper Functions
# NOTE: keep in synch with CheckBom function
BOM = { 'utf-8'  : codecs.BOM_UTF8,
        'utf-16' : codecs.BOM,
        'utf-32' : codecs.BOM_UTF32 }

# Regex for extracting magic comments from source files
# i.e *-* coding: utf-8 *-*, encoding=utf-8, ect...
# The first group from this expression will be the encoding.
RE_MAGIC_COMMENT = re.compile("coding[:=]\s*\"*([-\w.]+)\"*")

# File Load States
FL_STATE_START   = 0
FL_STATE_READING = 1
FL_STATE_PAUSED  = 2
FL_STATE_END     = 3
FL_STATE_ABORTED = 4

#--------------------------------------------------------------------------#

class ReadError(Exception):
    """Error happened while trying to read the file"""
    pass

class WriteError(Exception):
    """Error happened while trying to write the file"""
    pass

#--------------------------------------------------------------------------#

class EdFile(ebmlib.FileObjectImpl):
    """Wrapper for representing a file object that stores data
    about the file encoding and path.

    """
    _Checker = ebmlib.FileTypeChecker()

    def __init__(self, path=u'', modtime=0):
        """Create the file wrapper object
        @keyword path: the absolute path to the file
        @keyword modtime: file modification time

        """
        super(EdFile, self).__init__(path, modtime)

        # Attributes
        self._magic = dict(comment=None, bad=False)
        self.encoding = None
        self.bom = None
        self._mcallback = list()
        self.__buffer = None
        self._raw = False           # Raw bytes?
        self._fuzzy_enc = False
        self._job = None # async file read job

    def _SanitizeBOM(self, bstring):
        """Remove byte order marks that get automatically added by some codecs"""
        for enc in ('utf-8', 'utf-32', 'utf-16'):
            bmark = BOM.get(enc)
            if bstring.startswith(bmark):
                bstring = bstring.lstrip(bmark)
                break
        return bstring

    def _HandleRawBytes(self, bytes_value):
        """Handle prepping raw bytes for return to the buffer
        @param bytes_value: raw read bytes
        @return: string

        """
        Log("[ed_txt][info] HandleRawBytes called")
        if self._magic['comment']:
            self._magic['bad'] = True
        # Return the raw bytes to put into the buffer
        self._raw = True
        return '\0'.join(bytes_value)+'\0'

    def _ResetBuffer(self):
        Log("[ed_txt][info] Resetting buffer")
        if self.__buffer is not None:
            self.__buffer.close()
            del self.__buffer
        self.__buffer = StringIO()

    Encoding = property(lambda self: self.GetEncoding())

    def AddModifiedCallback(self, callback):
        """Set modified callback method
        @param callback: callable

        """
        self._mcallback.append(callback)

    def CleanUp(self):
        """Cleanup callback"""
        pass

    def Clone(self):
        """Clone the file object
        @return: EdFile

        """
        fileobj = EdFile(self.Path, self.ModTime)
        fileobj.SetLastError(self.last_err)
        fileobj.SetEncoding(self.encoding)
        fileobj.bom = self.bom
        fileobj._magic = dict(self._magic)
        fileobj._fuzzy_enc = self._fuzzy_enc
        for cback in self._mcallback:
            fileobj.AddModifiedCallback(cback)
        return fileobj

    def DecodeText(self):
        """Decode the text in the buffer and return a unicode string.
        @return: unicode or str

        """
        assert self.__buffer is not None, "No buffer!"
        assert self.encoding is not None, "Encoding Not Set!"

        bytes_value = self.__buffer.getvalue()
        ustr = u""
        try:
            if not self._fuzzy_enc or not EdFile._Checker.IsBinaryBytes(bytes_value):
                if self.bom is not None:
                    Log("[ed_txt][info] Stripping %s BOM from text" % self.encoding)
                    bytes_value = bytes_value.replace(self.bom, '', 1)

                Log("[ed_txt][info] Attempting to decode with: %s" % self.encoding)
                ustr = bytes_value.decode(self.encoding)
                # TODO: temporary...maybe
                # Check for utf-16 encodings which use double bytes
                # can result in NULLs in the string if decoded with
                # other encodings.
                if str('\0') in ustr:
                    Log("[ed_txt][info] NULL terminators found in decoded str")
                    Log("[ed_txt][info] Attempting UTF-16/32 detection...")
                    for utf_encoding in ('utf_16', 'utf_32'):
                        try:
                            tmpstr = bytes_value.decode(utf_encoding)
                        except UnicodeDecodeError:
                            pass
                        else:
                            self.encoding = utf_encoding
                            ustr = tmpstr
                            Log("[ed_txt][info] %s detected" % utf_encoding)
                            break
                    else:
                        Log("[ed_txt][info] No valid UTF-16/32 bytes detected")
            else:
                # Binary data was read
                Log("[ed_txt][info] Binary bytes where read")
                ustr = self._HandleRawBytes(bytes_value)
        except (UnicodeDecodeError, LookupError), msg:
            Log("[ed_txt][err] Error while reading with %s" % self.encoding)
            Log("[ed_txt][err] %s" % msg)
            self.SetLastError(unicode(msg))
            self.Close()
            # Decoding failed so convert to raw bytes for display
            ustr = self._HandleRawBytes(bytes_value)
        else:
            # Log success
            if not self._raw:
                Log("[ed_txt][info] Decoded %s with %s" % \
                    (self.GetPath(), self.encoding))

        # Scintilla bug, SetText will quit at first null found in the
        # string. So join the raw bytes and stuff them in the buffer instead.
        # TODO: are there other control characters that need to be checked
        #       for besides NUL?
        if not self._raw and '\0' in ustr:
            # Return the raw bytes to put into the buffer
            Log("[ed_txt][info] DecodeText - joining nul terminators")
            ustr = '\0'.join(bytes_value)+'\0'
            self._raw = True

        if self._raw:
            # TODO: wx/Scintilla Bug?
            # Replace \x05 with a space as it causes the buffer
            # to crash when its inserted.
            Log("[ed_txt][info] DecodeText - raw - set encoding to binary")
            ustr = ustr.replace('\x05', ' ')
            self.SetEncoding('binary')
            self._raw = True

        return ustr

    def DetectEncoding(self):
        """Try to determine the files encoding
        @precondition: File handle has been opened and is valid
        @postcondition: encoding and bom attributes will be set

        """
        if self.encoding != None:
            msg = ("[ed_txt][info] DetectEncoding, skipping do to user set "
                   "encoding: %s") % self.encoding
            Log(msg)
            return

        assert self.Handle is not None, "File handle not initialized"
        lines = [ self.Handle.readline() for x in range(2) ]
        self.Handle.seek(0)
        enc = None
        if len(lines):
            # First check for a Byte Order Mark
            enc = CheckBom(lines[0])

            # If no byte-order mark check for an encoding comment
            if enc is None:
                Log("[ed_txt][info] DetectEncoding - Check magic comment")
                self.bom = None
                if not self._magic['bad']:
                    enc = CheckMagicComment(lines)
                    if enc:
                        self._magic['comment'] = enc
            else:
                Log("[ed_txt][info] File Has %s BOM" % enc)
                self.bom = BOM.get(enc, None)

        if enc is None:
            Log("[ed_txt][info] Doing brute force encoding check")
            enc = GuessEncoding(self.GetPath(), 4096)

        if enc is None:
            self._fuzzy_enc = True
            enc = Profile_Get('ENCODING', default=DEFAULT_ENCODING)

        Log("[ed_txt][info] DetectEncoding - Set Encoding to %s" % enc)
        self.encoding = enc 

    def EncodeText(self):
        """Do a trial run of encoding all the text to ensure that we can
        determine an encoding that will correctly be able to write the data
        to disk.
        @return: bool

        """
        bOk = True
        txt = self.__buffer.read(8)
        self.__buffer.seek(0)
        if not ebmlib.IsUnicode(txt):
            # Already a string so nothing to do
            return bOk

        encs = GetEncodings()
        if self.encoding is None:
            self.encoding = Profile_Get('ENCODING', default=DEFAULT_ENCODING)
        encs.insert(0, self.encoding)
        cenc = self.encoding

        readsize = min(self.__buffer.len, 4096)
        for enc in encs:
            self.__buffer.seek(0)
            try:
                tmpchars = self.__buffer.read(readsize)
                while len(tmpchars):
                    tmpchars.encode(enc)
                    tmpchars = self.__buffer.read(readsize)
                self.encoding = enc
                self.__buffer.seek(0)
                self.ClearLastError()
            except LookupError, msg:
                Log("[ed_txt][err] Invalid encoding: %s" % enc)
                Log("[ed_txt][err] %s" % msg)
                self.SetLastError(unicode(msg))
            except UnicodeEncodeError, msg:
                Log("[ed_txt][err] Failed to encode text with %s" % enc)
                Log("[ed_txt][err] %s" % msg)
                self.SetLastError(unicode(msg))
            else:
                break
        else:
            bOk = False
            raise WriteError("Failed to encode text to byte string")

        # Log if the encoding changed due to encoding errors
        if self.encoding != cenc:
            Log("[ed_txt][warn] Used encoding %s differs from original %s" %\
                (self.encoding, cenc))

        self.__buffer.seek(0)
        return bOk

    def FireModified(self):
        """Fire the modified callback(s)"""
        remove = list()
        for idx, mcallback in enumerate(self._mcallback):
            try:
                mcallback()
            except:
                remove.append(idx)

        # Cleanup any bad callbacks
        if len(remove):
            remove.reverse()
            for idx in remove:
                self._mcallback.pop(idx)

    def GetEncoding(self):
        """Get the encoding used by the file it may not be the
        same as the encoding requested at construction time
        @return: string encoding name

        """
        if self.encoding is None:
            # Guard against early entry
            return Profile_Get('ENCODING', default=DEFAULT_ENCODING)
        return self.encoding

    def GetMagic(self):
        """Get the magic comment if one was present
        @return: string or None

        """
        return self._magic['comment']

    def HasBom(self):
        """Return whether the file has a bom byte or not
        @return: bool

        """
        return self.bom is not None

    def IsRawBytes(self):
        """Were only raw bytes read during the last read operation?
        @return: bool

        """
        return self._raw

    def IsReadOnly(self):
        """Return as read only when file is read only or if raw bytes"""
        return super(EdFile, self).IsReadOnly() or self.IsRawBytes()

    def Read(self, chunk=512):
        """Get the contents of the file as a string, automatically handling
        any decoding that may be needed.
        @keyword chunk: read size
        @return: unicode str
        @throws: ReadError Failed to open file for reading

        """
        if self.DoOpen('rb'):
            self.DetectEncoding()

            if self.encoding is None:
                # fall back to user setting
                self.encoding = Profile_Get('ENCODING', default=DEFAULT_ENCODING)
                Log(("[ed_txt][warn] Failed to detect encoding "
                    "falling back to default: %s") % self.encoding)

            self._ResetBuffer()
            self._raw = False

            Log("[ed_txt][info] Read - Start reading")
            tmp = self.Handle.read(chunk)
            while len(tmp):
                self.__buffer.write(tmp)
                tmp = self.Handle.read(chunk)
            Log("[ed_txt][info] Read - End reading")

            self.Close()
            txt = self.DecodeText()
            self.SetModTime(ebmlib.GetFileModTime(self.GetPath()))
            self._ResetBuffer()
            return txt
        else:
            Log("[ed_txt][err] Read Error: %s" % self.GetLastError())
            raise ReadError, self.GetLastError()

    def ReadAsync(self, control):
        """Read the file asynchronously on a separate thread
        @param control: text control to send text to

        """
        Log("[ed_txt][info] EdFile.ReadAsync()")
        pid = control.GetTopLevelParent().Id
        filesize = ebmlib.GetFileSize(self.GetPath())
        ed_msg.PostMessage(ed_msg.EDMSG_PROGRESS_STATE, (pid, 1, filesize))
        # Fork off async job to threadpool
        self._job = FileReadJob(control, self.ReadGenerator, 4096)
        ed_thread.EdThreadPool().QueueJob(self._job.run)

    def ReadGenerator(self, chunk=512):
        """Get the contents of the file as a string, automatically handling
        any decoding that may be needed.

        @keyword chunk: read size
        @return: unicode (generator)
        @throws: ReadError Failed to open file for reading.

        """
        if self.DoOpen('rb'):
            # Throttle yielded text to reduce event over head
            filesize = ebmlib.GetFileSize(self.Path)
            throttle = max(chunk, filesize/100)

            self.DetectEncoding()
            try:
                # Must use codec reader to ensure correct number of
                # bytes are read in to be decoded.
                reader = codecs.getreader(self.Encoding)(self.Handle)
                buffered_data = StringIO()
                while True:
                    tmp = reader.read(chunk)
                    if not len(tmp):
                        if buffered_data.len:
                            yield buffered_data.getvalue()
                            buffered_data.close()
                        break

                    buffered_data.write(tmp)
                    if buffered_data.len >= throttle:
                        yield buffered_data.getvalue()
                        buffered_data.close()
                        buffered_data = StringIO()
            except Exception, msg:
                Log("[ed_txt][err] Error while reading with %s" % self.Encoding)
                Log("[ed_txt][err] %s" % msg)
                self.SetLastError(unicode(msg))
                self.Close()
                if self._magic['comment']:
                    self._magic['bad'] = True

            Log("[ed_txt][info] Decoded %s with %s" % (self.Path, self.Encoding))
            self.SetModTime(ebmlib.GetFileModTime(self.Path))
        else:
            raise ReadError, self.GetLastError()

    def RemoveModifiedCallback(self, callback):
        """Remove a registered callback
        @param callback: callable to remove

        """
        if callback in self._mcallback:
            self._mcallback.remove(callback)

    def ResetAll(self):
        """Reset all attributes of this file"""
        super(EdFile, self).ResetAll()
        self._ResetBuffer()
        self._magic = dict(comment=None, bad=False)
        self.encoding = Profile_Get('ENCODING', default=DEFAULT_ENCODING)
        self.bom = None

    def SetEncoding(self, enc):
        """Explicitly set/change the encoding of the file
        @param enc: encoding to change to

        """
        if enc is None:
            enc = DEFAULT_ENCODING
        self.encoding = enc

    def ReadLines(self):
        """Get the contents of the file as a list of lines
        @return: list of strings

        """
        raise NotImplementedError

    def Write(self, value):
        """Write the given value to the file
        @param value: (Unicode) String of text to write to disk
        @note: exceptions are allowed to be raised for the writing
        @throws: WriteError Failed to open file for writing
        @throws: UnicodeEncodeError Failed to encode text using set encoding

        """
        ctime = time.time()
        Log("[ed_txt][info] Write - Called: %s - Time: %d" % (self.Path, ctime))

        # Check if a magic comment was added or changed
        self._ResetBuffer()
        self.__buffer.write(value)
        self.__buffer.seek(0)
        enc = CheckMagicComment([ self.__buffer.readline() for x in range(2) ])
        self.__buffer.seek(0)

        # Update encoding if necessary
        if enc is not None:
            Log("[ed_txt][info] Write: found magic comment: %s" % enc)
            self.encoding = enc

        # Encode to byte string
        # Do before opening file so that encoding failures don't cause file
        # data to get lost!
        if self.EncodeText():
            Log("[ed_txt][info] Write Successful test encode with %s" % self.Encoding)

            # Open and write the file
            if self.DoOpen('wb'):
                Log("[ed_txt][info] Opened %s, writing as %s" % (self.Path, self.Encoding))
                
                if self.HasBom():
                    Log("[ed_txt][info] Adding BOM back to text")
                    self.Handle.write(self.bom)

                # Write the file to disk
                chunk = min(self.__buffer.len, 4096)
                buffer_read = self.__buffer.read
                filewrite = self.Handle.write
                fileflush = self.Handle.flush
                sanitize = self._SanitizeBOM
                tmp = buffer_read(chunk)
                while len(tmp):
                    tmp_bytes = sanitize(tmp.encode(self.Encoding))
                    filewrite(tmp_bytes)
                    fileflush()
                    tmp = buffer_read(chunk)

                self._ResetBuffer() # Free buffer
                self.Close()
                Log("[ed_txt][info] %s was written successfully" % self.Path)
            else:
                self._ResetBuffer()
                raise WriteError, self.GetLastError()

        Log("[ed_txt][info] Write - Complete: %s - Time: %d" % 
            (self.Path, time.time() - ctime))

#-----------------------------------------------------------------------------#

class FileReadJob(object):
    """Job for running an async file read in a background thread"""
    def __init__(self, receiver, task, *args, **kwargs):
        """Create the thread
        @param receiver: Window to receive events
        @param task: generator method to call
        @param *args: positional arguments to pass to task
        @param **kwargs: keyword arguments to pass to task

        """
        super(FileReadJob, self).__init__()

        # Attributes
        self.cancel = False
        self._task = task
        self.receiver = receiver
        self._args = args
        self._kwargs = kwargs
        self.pid = receiver.TopLevelParent.Id

    def run(self):
        """Read the text"""
        evt = FileLoadEvent(edEVT_FILE_LOAD, wx.ID_ANY, None, FL_STATE_START)
        wx.PostEvent(self.receiver, evt)
        time.sleep(.75) # give ui a chance to get ready

        count = 1
        for txt in self._task(*self._args, **self._kwargs):
            if self.cancel:
                break

            evt = FileLoadEvent(edEVT_FILE_LOAD, wx.ID_ANY, txt)
            evt.SetProgress(count * self._args[0])
            wx.PostEvent(self.receiver, evt)
            count += 1

        evt = FileLoadEvent(edEVT_FILE_LOAD, wx.ID_ANY, None, FL_STATE_END)
        wx.PostEvent(self.receiver, evt)

    def Cancel(self):
        """Cancel the running task"""
        self.cancel = True

#-----------------------------------------------------------------------------#

edEVT_FILE_LOAD = wx.NewEventType()
EVT_FILE_LOAD = wx.PyEventBinder(edEVT_FILE_LOAD, 1)
class FileLoadEvent(wx.PyEvent):
    """Event to signal that a chunk of text haes been read"""
    def __init__(self, etype, eid, value=None, state=FL_STATE_READING):
        """Creates the event object"""
        super(FileLoadEvent, self).__init__(eid, etype)

        # Attributes
        self._state = state
        self._value = value
        self._prog = 0
    
    def HasText(self):
        """Returns true if the event has text
        @return: bool whether the event contains text

        """
        return self._value is not None

    def GetProgress(self):
        """Get the current progress of the load"""
        return self._prog

    def GetState(self):
        """Get the state of the file load action
        @return: int (FL_STATE_FOO)

        """
        return self._state

    def GetValue(self):
        """Returns the value from the event.
        @return: the value of this event

        """
        return self._value

    def SetProgress(self, progress):
        """Set the number of bytes that have been read
        @param progress: int

        """
        self._prog = progress

#-----------------------------------------------------------------------------#
# Utility Function
def CheckBom(line):
    """Try to look for a bom byte at the beginning of the given line
    @param line: line (first line) of a file
    @return: encoding or None

    """
    Log("[ed_txt][info] CheckBom called")
    has_bom = None
    # NOTE: MUST check UTF-32 BEFORE utf-16
    for enc in ('utf-8', 'utf-32', 'utf-16'):
        bom = BOM[enc]
        if line.startswith(bom):
            has_bom = enc
            break
    return has_bom

def CheckMagicComment(lines):
    """Try to decode the given text on the basis of a magic
    comment if one is present.
    @param lines: list of lines to check for a magic comment
    @return: encoding or None

    """
    Log("[ed_txt][info] CheckMagicComment: %s" % str(lines))
    enc = None
    for line in lines:
        match = RE_MAGIC_COMMENT.search(line)
        if match:
            enc = match.group(1)
            try:
                codecs.lookup(enc)
            except LookupError:
                enc = None
            break

    Log("[ed_txt][info] MagicComment is %s" % enc)
    return enc

def DecodeString(string, encoding=None):
    """Decode the given string to Unicode using the provided
    encoding or the DEFAULT_ENCODING if None is provided.
    @param string: string to decode
    @keyword encoding: encoding to decode string with

    """
    if encoding is None:
        encoding = DEFAULT_ENCODING

    if not ebmlib.IsUnicode(string):
        try:
            rtxt = string.decode(encoding)
        except Exception, msg:
            Log("[ed_txt][err] DecodeString with %s failed" % encoding)
            Log("[ed_txt][err] %s" % msg)
            rtxt = string
        return rtxt
    else:
        # The string is already unicode so just return it
        return string

def EncodeString(string, encoding=None):
    """Try and encode a given unicode object to a string
    with the provided encoding returning that string. The
    default encoding will be used if None is given for the
    encoding.
    @param string: unicode object to encode into a string
    @keyword encoding: encoding to use for conversion

    """
    if not encoding:
        encoding = DEFAULT_ENCODING

    if ebmlib.IsUnicode(string):
        try:
            rtxt = string.encode(encoding)
        except LookupError:
            rtxt = string
        return rtxt
    else:
        return string

def FallbackReader(fname):
    """Guess the encoding of a file by brute force by trying one
    encoding after the next until something succeeds.
    @param fname: file path to read from
    @todo: deprecate this method

    """
    txt = None
    with open(fname, 'rb') as handle:
        byte_str = handle.read()
        for enc in GetEncodings():
            try:
                txt = byte_str.decode(enc)
            except Exception, msg:
                continue
            else:
                return (enc, txt)

    return (None, None)

def GuessEncoding(fname, sample):
    """Attempt to guess an encoding
    @param fname: filename
    @param sample: pre-read amount
    @return: encoding or None

    """
    for enc in GetEncodings():
        try:
            with open(fname, 'rb') as handle:
                with codecs.getreader(enc)(handle) as reader:
                    value = reader.read(sample)
                    if str('\0') in value:
                        continue
                    else:
                        return enc
        except Exception, msg:
            continue
    return None

def GetEncodings():
    """Get a list of possible encodings to try from the locale information
    @return: list of strings

    """
    encodings = list()
    encodings.append(Profile_Get('ENCODING', None))

    try:
        encodings.append(locale.getpreferredencoding())
    except:
        pass
    
    encodings.append('utf-8')

    try:
        if hasattr(locale, 'nl_langinfo'):
            encodings.append(locale.nl_langinfo(locale.CODESET))
    except:
        pass
    try:
        encodings.append(locale.getlocale()[1])
    except:
        pass
    try:
        encodings.append(locale.getdefaultlocale()[1])
    except:
        pass
    encodings.append(sys.getfilesystemencoding())
    encodings.append('utf-16')
    encodings.append('utf-16-le') # for files without BOM...
    encodings.append('latin-1')

    # Normalize all names
    normlist = [ enclib.normalize_encoding(enc) for enc in encodings if enc]

    # Clean the list for duplicates and None values
    rlist = list()
    codec_list = list()
    for enc in normlist:
        if enc is not None and len(enc):
            enc = enc.lower()
            if enc not in rlist:
                # Ascii is useless so ignore it (ascii, us_ascii, ...)
                if 'ascii' in enc:
                    continue

                try:
                    ctmp = codecs.lookup(enc)
                    if ctmp.name not in codec_list:
                        codec_list.append(ctmp.name)
                        rlist.append(enc)
                except LookupError:
                    pass
    return rlist



###############################################################################
# Name: OutputBufferDemo.py                                                   #
# Purpose: OutputBuffer Test and Demo File                                    #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Test file for testing the OutputBuffer (eclib.eclib) module and controls. This
demo/test contains a small application for running 'ping' and other command line
commands and then displaying their output. Special text styling is provided for
ping to highlight and create hotspot regions on IP and website addresses.

Clicking the start button multiple times will spawn multiple processes and
threads to interact with the buffer.


"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: OutputBufferDemo.py 63510 2010-02-18 02:30:38Z CJP $"
__revision__ = "$Revision: 63510 $"

#-----------------------------------------------------------------------------#
# Imports
import os
import sys
import re
import wx

#sys.path.insert(0, os.path.abspath('../../src'))
import eclib

#-----------------------------------------------------------------------------#
# Globals

ID_COMMAND = wx.NewId()
ID_START = wx.NewId()
ID_STOP = wx.NewId()
ID_PROC_LIST = wx.NewId()
ID_THREAD_STATUS = wx.NewId()
#-----------------------------------------------------------------------------#

class TestPanel(wx.Panel):
    def __init__(self, parent, log):
        self.log = log
        wx.Panel.__init__(self, parent)

        # Attributes
        self.ssizer = wx.BoxSizer(wx.HORIZONTAL)
        self._buff = ProcessOutputBuffer(self, self.log)

        # Layout
        self.__DoLayout()

        # Event Handlers
        self.Bind(wx.EVT_BUTTON, self.OnButton)

    def __DoLayout(self):
        """Layout the panel"""

        # Status Display Layout (Top Section)
        self.ssizer.AddMany([((5, 5),),
                             (wx.StaticText(self, label="Running:"),
                              0, wx.ALIGN_CENTER_VERTICAL),
                             (wx.Choice(self, ID_PROC_LIST), 0),
                             ((-1, 5), 1, wx.EXPAND),
                             (wx.StaticText(self, ID_THREAD_STATUS, "Threads: 0"),
                              0, wx.ALIGN_CENTER_VERTICAL),
                             ((5, 5),)])

        default = ["ping localhost", "ping editra.org", "ping wxpython.org"]
        combo = wx.ComboBox(self, ID_COMMAND, value='ping localhost', choices=default)
        combo.SetToolTipString("Enter a command to run here")
        hsizer = wx.BoxSizer(wx.HORIZONTAL)
        stopb = wx.Button(self, ID_STOP, "Stop")
        stopb.SetToolTipString("Stop all running processes")
        startb = wx.Button(self, ID_START, "Start")
        startb.SetToolTipString("Run the command in the combo box\n"
                                "Press multiple times to start multiple threads")
        hsizer.AddMany([((5, 5),), (wx.StaticText(self, label="Cmd: "), 0,
                                    wx.ALIGN_CENTER_VERTICAL|wx.ALIGN_LEFT),
                        (combo, 1, wx.EXPAND), ((20, 20),),
                        (stopb, 0, wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL),
                        ((5, 5),),
                        (startb, 0, wx.ALIGN_RIGHT|wx.ALIGN_CENTER_VERTICAL),
                        ((5, 5),)])

        # Main Sizer Layout
        msizer = wx.BoxSizer(wx.VERTICAL)
        msizer.AddMany([((3, 3),),
                        (self.ssizer, 0, wx.EXPAND),
                        (self._buff, 1, wx.EXPAND),
                        (hsizer, 0, wx.EXPAND)])
        self.SetSizer(msizer)

    def OnButton(self, evt):
        """Start/Stop ProcessThread(s)"""
        e_id = evt.GetId()
        if e_id == ID_STOP:
            self._buff.Abort()
            self.UpdateProcs()
        elif e_id == ID_START:
            # Spawn a new ProcessThread
            combo = self.FindWindowById(ID_COMMAND)
            self._buff.StartProcess(u'%s' % combo.GetValue())
            self.UpdateProcs()
        else:
            evt.Skip()

    def UpdateProcs(self):
        """Update the controls to show the current processes/threads"""
        procs = self.FindWindowById(ID_PROC_LIST)
        status = self.FindWindowById(ID_THREAD_STATUS)

        running = self._buff.GetProcesses()
        procs.SetItems(running)
        if len(running):
            procs.SetStringSelection(running[-1])
        status.SetLabel("Threads: %d" % len(running))
        self.ssizer.Layout()

#-----------------------------------------------------------------------------#
STARTED_STR = 'ProcessThread Started'
FINISHED_STR = 'ProcessThread Finished'

class ProcessOutputBuffer(eclib.OutputBuffer, eclib.ProcessBufferMixin):
    """Custom output buffer for processing output from running ping"""
    RE_ADDR = re.compile(r' ([a-z0-9]+\.[a-z0-9]+\.{0,1})+')
    ADDRESS_STYLE = eclib.OPB_STYLE_MAX + 1

    def __init__(self, parent, log):
        eclib.OutputBuffer.__init__(self, parent)
        eclib.ProcessBufferMixin.__init__(self, update=125)

        # Attributes
        self.log = log
        self._threads = list()

        # Setup a custom style for highlighting IP addresses
        font = self.GetFont()
        style = (font.GetFaceName(), font.GetPointSize(), "#FFFFFF")
        self.StyleSetSpec(ProcessOutputBuffer.ADDRESS_STYLE, 
                          "face:%s,size:%d,fore:#3030FF,back:%s" % style)
        self.StyleSetHotSpot(ProcessOutputBuffer.ADDRESS_STYLE, True)

    def Abort(self):
        """Kill off all running threads and procesess"""
        for proc in self._threads:
            proc[0].Abort()

    def ApplyStyles(self, ind, txt):
        """Highlight the begining and process exit text in the buffer
        Tests overriding of the ApplyStyles method from L{eclib.OutputBuffer}
        This is called every time text is added to the output buffer.
        @param ind: Index of text insertion point in buffer
        @param txt: the text that was just inserted at ind

        """
        # Highlight ip addresses
        for group in ProcessOutputBuffer.RE_ADDR.finditer(txt):
            sty_s = ind + group.start() + 1 # start past whitespace
            sty_e = ind + group.end()
            self.StartStyling(sty_s, 0xff)
            self.SetStyling(sty_e - sty_s, ProcessOutputBuffer.ADDRESS_STYLE)

        # Check for first/last messages
        start = txt.find(STARTED_STR)
        end = txt.find(FINISHED_STR)
        if start >= 0:
            sty_s = ind + start
            slen = len(STARTED_STR)
            style = eclib.OPB_STYLE_INFO
        elif end >= 0:
            sty_s = ind + end
            slen = len(FINISHED_STR)
            style = eclib.OPB_STYLE_WARN
        else:
            return

        # Do the styling if one of the patterns was found
        self.StartStyling(sty_s, 0xff)
        self.SetStyling(slen, style)

    def DoHotSpotClicked(self, pos, line):
        """Overridden method from base L{eclib.OutputBuffer} class.
        This method is called whenever a hotspot in the output buffer is
        clicked on.
        @param pos: Click postion in buffer
        @param line: Line click occurred on

        """
        self.log.write("Hotspot Clicked: Pos %d, Line %d" % (pos, line))
        if self.GetStyleAt(pos) == ProcessOutputBuffer.ADDRESS_STYLE:
            ipaddr = ProcessOutputBuffer.RE_ADDR.search(self.GetLine(line))
            if ipaddr is not None:
                ipaddr = ipaddr.group(0)
                self.log.write("Address Clicked: %s" % ipaddr)

    def DoProcessStart(self, cmd=''):
        """Do any necessary preprocessing before a process is started.
        This method is called directly before the process in the
        L{eclib.ProcessThread} is started. This method is an overridden 
        method of the L{eclib.ProcessBufferMixin} super class.
        @keyword cmd: Command string used to start the process

        """
        self.AppendUpdate("%s: %s%s" % (STARTED_STR, cmd, os.linesep))

    def DoProcessExit(self, code=0):
        """Do all that is needed to be done after a process has exited
        This method is called when a L{eclib.ProcessThread} has exited and
        is an overridden method from the L{eclib.ProcessBufferMixin} super
        class.
        @keyword code: Exit code of the process

        """
        self.AppendUpdate("%s: %d%s" % (FINISHED_STR, code, os.linesep))
        dead = list()

        # Remove all Dead Threads
        for idx, proc in enumerate(self._threads):
            if not proc[0].isAlive():
                dead.append(idx)

        for zombie in reversed(dead):
            del self._threads[zombie]

        self.GetParent().UpdateProcs()

        # If there are no more running threads stop the buffers timer
        if not len(self._threads):
            self.Stop()

    def GetActiveCount(self):
        """Get the count of active threaded processes that are running
        @return: int

        """
        return len(self._threads)

    def GetProcesses(self):
        """Return the list of running process commands
        @return: list of strings

        """
        return [ proc[1] for proc in self._threads ]

    def StartProcess(self, cmd):
        """Start a new process thread, the process will be created by
        executing the given command.
        @param cmd: Command string (i.e 'ping localhost')

        """
        proc = eclib.ProcessThread(self, cmd, env=os.environ)
        self._threads.append((proc, cmd))
        self._threads[-1][0].start()
        
#-----------------------------------------------------------------------------#

def runTest(frame, nb, log):
    return TestPanel(nb, log)

class TestLog:
    def __init__(self):
        pass

    def write(self, msg):
        print msg

#----------------------------------------------------------------------

overview = eclib.outbuff.__doc__
title = "OutputBuffer"

#-----------------------------------------------------------------------------#
# Run the Test
if __name__ == '__main__':
    try:
        import run
    except ImportError:
        app = wx.PySimpleApp(False)
        frame = wx.Frame(None, title="OutputBuffer Test 'Ping-O-Matic'")
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(TestPanel(frame, TestLog()), 1, wx.EXPAND)
        frame.CreateStatusBar()
        frame.SetSizer(sizer)
        frame.SetInitialSize((500, 500))
        frame.SetStatusText("OutputBufferDemo: wxPython %s, Python %s" % \
                            (wx.__version__, u".".join(str(x) for x in sys.version_info)))
        frame.CenterOnParent()
        frame.Show()
        app.MainLoop()
    else:
        run.main(['', os.path.basename(sys.argv[0])] + sys.argv[1:])



###############################################################################
# Name: haskell.py                                                            #
# Purpose: Define Haskell syntax for highlighting and other features          #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
 FILE: haskell.py
 AUTHOR: Cody Precord
 @summary: Lexer configuration module for the Haskell Programming Language
 @todo: More custom highlighting

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _haskell.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import synglob
import syndata

#-----------------------------------------------------------------------------#

#---- Keyword Definitions ----#
HA_KEYWORDS = (0, "as case class data default deriving do forall foreign "
                  "hiding if import in infix infixl infixr instance else let "
                  "mdo module newtype of qualified then type where")

#---- End Keyword Definitions ----#

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [(stc.STC_HA_CAPITAL, 'default_style'),
                (stc.STC_HA_CHARACTER, 'char_style'),
                (stc.STC_HA_CLASS, 'class_style'),
                (stc.STC_HA_COMMENTBLOCK, 'comment_style'),
                (stc.STC_HA_COMMENTBLOCK2, 'comment_style'),
                (stc.STC_HA_COMMENTBLOCK3, 'comment_style'),
                (stc.STC_HA_COMMENTLINE, 'comment_style'),
                (stc.STC_HA_DATA, 'default_style'),
                (stc.STC_HA_DEFAULT, 'default_style'),
                (stc.STC_HA_IDENTIFIER, 'default_style'),
                (stc.STC_HA_IMPORT, 'default_style'), # possibly use custom style
                (stc.STC_HA_INSTANCE, 'default_style'),
                (stc.STC_HA_KEYWORD, 'keyword_style'),
                (stc.STC_HA_MODULE, 'default_style'),
                (stc.STC_HA_NUMBER, 'number_style'),
                (stc.STC_HA_OPERATOR, 'operator_style'),
                (stc.STC_HA_STRING, 'string_style')]

#---- Extra Properties ----#
FOLD = ('fold', '1')

#-----------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Haskell""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_HASKELL)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        return [HA_KEYWORDS]

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [FOLD]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code """
        return [u'--']



###############################################################################
# Name: ctrlbox.py                                                            #
# Purpose: Container Window helper class                                      #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2008 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Editra Control Library: ControlBox

Sizer managed panel class with support for a toolbar like control that can be
placed on the top/bottom of the main window area, multiple control bars are also
possible.

Class ControlBar:

Toolbar like control with automatic item spacing and layout.

Styles:
  - CTRLBAR_STYLE_DEFAULT: Plain background
  - CTRLBAR_STYLE_GRADIENT: Draw the bar with a vertical gradient.
  - CTRLBAR_STYLE_BORDER_BOTTOM: add a border to the bottom
  - CTRLBAR_STYLE_BORDER: add a border to the top
  - CTRLBAR_STYLE_VERTICAL = Vertical ControlBar tool layout

Class ControlBox:

The ControlBox is a sizer managed panel that supports easy creation of windows
that require a sandwich like layout.

+---------------------------------------+
| ControlBar                            |
+---------------------------------------+
|                                       |
|                                       |
|          MainWindow Area              |
|                                       |
|                                       |
|                                       |
+---------------------------------------+
| ControlBar                            |
+---------------------------------------+

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: ctrlbox.py 71187 2012-04-11 23:38:30Z CJP $"
__revision__ = "$Revision: 71187 $"

__all__ = ["ControlBox", "CTRLBOX_NAME_STR",

           "ControlBar", "ControlBarEvent",
           "CTRLBAR_STYLE_DEFAULT", "CTRLBAR_STYLE_GRADIENT",
           "CTRLBAR_STYLE_BORDER_TOP", "CTRLBAR_STYLE_BORDER_BOTTOM",
           "CTRLBAR_STYLE_VERTICAL",
           "EVT_CTRLBAR", "edEVT_CTRLBAR", "CTRLBAR_NAME_STR",

           "SegmentBar", "SegmentBarEvent",
           "EVT_SEGMENT_SELECTED", "edEVT_SEGMENT_SELECTED",
           "EVT_SEGMENT_CLOSE", "edEVT_SEGMENT_CLOSE",
           "CTRLBAR_STYLE_LABELS", "CTRLBAR_STYLE_NO_DIVIDERS",
           "SEGBTN_OPT_CLOSEBTNL", "SEGBTN_OPT_CLOSEBTNR",
           "SEGBAR_NAME_STR", "SEGMENT_HT_NOWHERE",
           "SEGMENT_HT_SEG", "SEGMENT_HT_X_BTN"
]

#--------------------------------------------------------------------------#
# Dependencies
import wx

# Local Imports
from eclutil import AdjustColour, DrawCircleCloseBmp

#--------------------------------------------------------------------------#
# Globals

#-- Control Name Strings --#
CTRLBAR_NAME_STR = u'EditraControlBar'
CTRLBOX_NAME_STR = u'EditraControlBox'
SEGBAR_NAME_STR = u'EditraSegmentBar'

#-- Control Style Flags --#

# ControlBar / SegmentBar Style Flags
CTRLBAR_STYLE_DEFAULT       = 0
CTRLBAR_STYLE_GRADIENT      = 1     # Paint the bar with a gradient
CTRLBAR_STYLE_BORDER_BOTTOM = 2     # Add a border to the bottom
CTRLBAR_STYLE_BORDER_TOP    = 4     # Add a border to the top
CTRLBAR_STYLE_LABELS        = 8     # Draw labels under the icons (SegmentBar)
CTRLBAR_STYLE_NO_DIVIDERS   = 16    # Don't draw dividers between segments
CTRLBAR_STYLE_VERTICAL      = 32    # Control bar in vertical orientation

# Segment Button Options
SEGBTN_OPT_NONE          = 1     # No options set.
SEGBTN_OPT_CLOSEBTNL     = 2     # Close button on the segments left side.
SEGBTN_OPT_CLOSEBTNR     = 4     # Close button on the segment right side.

# Hit test locations
SEGMENT_HT_NOWHERE = 0
SEGMENT_HT_SEG     = 1
SEGMENT_HT_X_BTN   = 2

# Segment States
SEGMENT_STATE_NONE = 0  # Hover no where
SEGMENT_STATE_SEG  = 1  # Hover on segment
SEGMENT_STATE_X    = 2  # Hover on segment x button

# ControlBar event for items added by AddTool
edEVT_CTRLBAR = wx.NewEventType()
EVT_CTRLBAR = wx.PyEventBinder(edEVT_CTRLBAR, 1)
class ControlBarEvent(wx.PyCommandEvent):
    """ControlBar Button Event"""

edEVT_SEGMENT_SELECTED = wx.NewEventType()
EVT_SEGMENT_SELECTED = wx.PyEventBinder(edEVT_SEGMENT_SELECTED, 1)
edEVT_SEGMENT_CLOSE = wx.NewEventType()
EVT_SEGMENT_CLOSE = wx.PyEventBinder(edEVT_SEGMENT_CLOSE, 1)
class SegmentBarEvent(wx.PyCommandEvent):
    """SegmentBar Button Event"""
    def __init__(self, etype, id=0):
        super(SegmentBarEvent, self).__init__(etype, id)

        # Attributes
        self.notify = wx.NotifyEvent(etype, id)
        self._pre = -1
        self._cur = -1

    def GetPreviousSelection(self):
        """Get the previously selected segment
        @return: int

        """
        return self._pre

    def GetCurrentSelection(self):
        """Get the currently selected segment
        @return: int

        """
        return self._cur

    def IsAllowed(self):
        """Is the event allowed to propagate
        @return: bool

        """
        return self.notify.IsAllowed()

    def SetSelections(self, previous=-1, current=-1):
        """Set the events selection
        @keyword previous: previously selected button index (int)
        @keyword current: currently selected button index (int)

        """
        self._pre = previous
        self._cur = current

    def Veto(self):
        """Veto the event"""
        self.notify.Veto()

#--------------------------------------------------------------------------#

class ControlBox(wx.PyPanel):
    """Simple managed panel helper class that allows for adding and
    managing the position of a small toolbar like panel.
    @see: L{ControlBar}

    """
    def __init__(self, parent, id=wx.ID_ANY,
                 pos=wx.DefaultPosition, size=wx.DefaultSize,
                 style=wx.TAB_TRAVERSAL|wx.NO_BORDER,
                 name=CTRLBOX_NAME_STR):
        super(ControlBox, self).__init__(parent, id, pos, size, style, name)

        # Attributes
        self._vsizer = wx.BoxSizer(wx.VERTICAL)
        self._hsizer = wx.BoxSizer(wx.HORIZONTAL)
        self._cbars = dict()
        self._main = None

        # Layout
        self._hsizer.Add(self._vsizer, 1, wx.EXPAND)
        self.SetSizer(self._hsizer)

    #---- Properties ----#
    Window = property(lambda self: self.GetWindow())

    def _GetCtrlBarSizer(self, pos):
        """Get the correct sizer for the ControlBar at the given pos
        @param pos: wx.TOP/LEFT/RIGHT/BOTTOM

        """
        if pos in (wx.TOP, wx.BOTTOM):
            sizer = self._vsizer
        else:
            sizer = self._hsizer
        return sizer

    def ChangeWindow(self, window):
        """Change the main window area, and return the current window
        @param window: Any window/panel like object
        @return: the old window or None

        """
        rwindow = None
        if self.GetWindow() is None or not isinstance(self._main, wx.Window):
            del self._main
            topb = self.GetControlBar(wx.TOP)
            if topb is None:
                self._vsizer.Add(window, 1, wx.EXPAND)
            else:
                self._vsizer.Insert(1, window, 1, wx.EXPAND)
        else:
            self._vsizer.Replace(self._main, window)
            rwindow = self._main

        self._main = window
        return rwindow

    def CreateControlBar(self, pos=wx.TOP):
        """Create a ControlBar at the given position if one does not
        already exist.
        @keyword pos: wx.TOP (default), BOTTOM, LEFT, RIGHT
        @postcondition: A top aligned L{ControlBar} is created.
        @return: ControlBar

        """
        cbar = self.GetControlBar(pos)
        if cbar is None:
            style=CTRLBAR_STYLE_GRADIENT
            if pos in (wx.LEFT, wx.RIGHT):
                style |= CTRLBAR_STYLE_VERTICAL
            cbar = ControlBar(self, style=style)
            self.SetControlBar(cbar, pos)
        return cbar

    def GetControlBar(self, pos=wx.TOP):
        """Get the L{ControlBar} used by this window
        @param pos: wx.TOP, BOTTOM, LEFT, RIGHT
        @return: ControlBar or None

        """
        assert pos in (wx.TOP, wx.LEFT, wx.BOTTOM, wx.RIGHT)
        cbar = self._cbars.get(pos, None)
        return cbar

    def GetWindow(self):
        """Get the main display window
        @return: Window or None

        """
        return self._main

    def ReplaceControlBar(self, ctrlbar, pos=wx.TOP):
        """Replace the L{ControlBar} at the given position
        with the given ctrlbar and return the bar that was
        replaced or None.
        @param ctrlbar: L{ControlBar}
        @keyword pos: Position
        @return: L{ControlBar} or None

        """
        assert isinstance(ctrlbar, ControlBar)
        assert pos in (wx.TOP, wx.BOTTOM, wx.LEFT, wx.RIGHT)
        tbar = self.GetControlBar(pos)
        rbar = None
        sizer = self._GetCtrlBarSizer(pos)
        if tbar is None and pos in (wx.TOP, wx.LEFT):
            sizer.Insert(0, ctrlbar, 0, wx.EXPAND)
        elif tbar is None and pos in (wx.BOTTOM, wx.RIGHT):
            sizer.Add(ctrlbar, 0, wx.EXPAND)
        else:
            sizer.Replace(tbar, ctrlbar)
            rbar = tbar

        self._cbars[pos] = ctrlbar

        return rbar

    def SetControlBar(self, ctrlbar, pos=wx.TOP):
        """Set the ControlBar used by this ControlBox
        @param ctrlbar: L{ControlBar}
        @keyword pos: wx.TOP/wx.BOTTOM/wx.LEFT/wx.RIGHT

        """
        assert isinstance(ctrlbar, ControlBar)
        assert pos in (wx.TOP, wx.BOTTOM, wx.LEFT, wx.RIGHT)
        tbar = self.GetControlBar(pos)
        if tbar is ctrlbar:
            return # ignore setting same bar again
        sizer = self._GetCtrlBarSizer(pos)
        if tbar is None and pos in (wx.TOP, wx.LEFT):
            sizer.Insert(0, ctrlbar, 0, wx.EXPAND)
        elif tbar is None and pos in (wx.BOTTOM, wx.RIGHT):
            sizer.Add(ctrlbar, 0, wx.EXPAND)
        else:
            sizer.Replace(tbar, ctrlbar)

            try:
                tbar.Destroy()
            except wx.PyDeadObjectError:
                pass

        self._cbars[pos] = ctrlbar

    def SetWindow(self, window):
        """Set the main window control portion of the box. This will be the
        main central item shown in the box
        @param window: Any window/panel like object

        """
        if self.GetWindow() is None:
            topb = self.GetControlBar(wx.TOP)
            botb = self.GetControlBar(wx.BOTTOM)
            if (topb and botb is None) or (topb is None and botb is None):
                self._vsizer.Add(window, 1, wx.EXPAND)
            elif botb and topb is None:
                self._vsizer.Insert(0, window, 1, wx.EXPAND)
            else:
                self._vsizer.Insert(1, window, 1, wx.EXPAND)
        else:
            self._vsizer.Replace(self._main, window)

            try:
                self._main.Destroy()
            except wx.PyDeadObjectError:
                pass

        self._main = window

#--------------------------------------------------------------------------#

class ControlBar(wx.PyPanel):
    """Toolbar like control container for use with a L{ControlBox}. It
    uses a panel with a managed sizer as a convenient way to add a small
    bar with various controls in it to any window.

    """
    def __init__(self, parent, id=wx.ID_ANY,
                 pos=wx.DefaultPosition, size=wx.DefaultSize,
                 style=CTRLBAR_STYLE_DEFAULT,
                 name=CTRLBAR_NAME_STR):
        super(ControlBar, self).__init__(parent, id, pos, size,
                                         wx.TAB_TRAVERSAL|wx.NO_BORDER, name)

        tsz_orient = wx.HORIZONTAL
        msz_orient = wx.VERTICAL
        if style & CTRLBAR_STYLE_VERTICAL:
            tsz_orient = wx.VERTICAL
            msz_orient = wx.HORIZONTAL

        # Attributes
        self._style = style
        self._sizer = wx.BoxSizer(tsz_orient)
        self._tools = dict(simple=list())
        self._spacing = (5, 5)

        # Drawing related
        color = wx.SystemSettings_GetColour(wx.SYS_COLOUR_3DFACE)
        if wx.Platform != '__WXMAC__':
            self._color2 = AdjustColour(color, 15)
            self._color = AdjustColour(color, -10)
        else:
            self._color2 = AdjustColour(color, 15)
            self._color = AdjustColour(color, -20)

        pcolor = tuple([min(190, x) for x in AdjustColour(self._color, -25)])
        self._pen = wx.Pen(pcolor, 1)

        # Setup
        msizer = wx.BoxSizer(msz_orient)
        spacer = (0, 0)
        msizer.Add(spacer, 0)
        msizer.Add(self._sizer, 1, wx.EXPAND)
        msizer.Add(spacer, 0)
        self.SetSizer(msizer)

        # Event Handlers
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_BUTTON, self._DispatchEvent)

    def _DispatchEvent(self, evt):
        """Translate the button events generated by the controls added by
        L{AddTool} to L{ControlBarEvent}'s.

        """
        e_id = evt.GetId()
        if e_id in self._tools['simple']:
            cb_evt = ControlBarEvent(edEVT_CTRLBAR, e_id)
            self.GetEventHandler().ProcessEvent(cb_evt)
        else:
            # Allow to propagate
            evt.Skip()

    def _GetAlignment(self):
        """Verify and get the proper secondary alignment based on the
        control bar alignment.

        """
        if not self.IsVerticalMode():
            align2 = wx.ALIGN_CENTER_VERTICAL
        else:
            align2 = wx.ALIGN_CENTER_HORIZONTAL
        return align2

    def AddControl(self, control, align=-1, stretch=0):
        """Add a control to the bar
        @param control: The control to add to the bar
        @keyword align: wx.ALIGN_**
        @keyword stretch: The controls proportions 0 for normal, 1 for expand

        """
        if wx.Platform == '__WXMAC__':
            if hasattr(control, 'SetWindowVariant'):
                control.SetWindowVariant(wx.WINDOW_VARIANT_SMALL)

        # Default to proper alignment when -1 specified
        if align not in (wx.ALIGN_LEFT, wx.ALIGN_RIGHT,
                         wx.ALIGN_BOTTOM, wx.ALIGN_TOP):
            if self.IsVerticalMode():
                align = wx.ALIGN_TOP
            else:
                align = wx.ALIGN_LEFT

        align2 = self._GetAlignment()
        if align in (wx.ALIGN_LEFT, wx.ALIGN_TOP):
            self._sizer.Add(self._spacing, 0)
            self._sizer.Add(control, stretch, align|align2)
        else:
            self._sizer.Add(control, stretch, align|align2)
            self._sizer.Add(self._spacing, 0)

        self.Layout()

    def AddSpacer(self, width, height):
        """Add a fixed size spacer to the control bar
        @param width: width of the spacer
        @param height: height of the spacer

        """
        self._sizer.Add((width, height), 0)

    def AddStretchSpacer(self):
        """Add an expanding spacer to the bar that will stretch and
        contract when the window changes size.

        """
        self._sizer.AddStretchSpacer(2)

    def AddTool(self, tid, bmp, help=u'', align=-1):
        """Add a simple bitmap button tool to the control bar
        @param tid: Tool Id
        @param bmp: Tool bitmap
        @keyword help: Short help string
        @keyword align: wx.ALIGN_**

        """
        tool = wx.BitmapButton(self, tid, bmp, style=wx.NO_BORDER)
        if wx.Platform == '__WXGTK__':
            # SetMargins not available in wxPython 2.9+
            getattr(tool, 'SetMargins', lambda x,y: False)(0, 0)
            spacer = (0, 0)
        else:
            spacer = self._spacing
        tool.SetToolTipString(help)

        # Default to proper alignment when unknown is specified
        if align not in (wx.ALIGN_LEFT, wx.ALIGN_RIGHT,
                         wx.ALIGN_BOTTOM, wx.ALIGN_TOP):
            if self.IsVerticalMode():
                align = wx.ALIGN_TOP
            else:
                align = wx.ALIGN_LEFT

        align2 = self._GetAlignment()
        self._tools['simple'].append(tool.GetId())
        if align in (wx.ALIGN_TOP, wx.ALIGN_LEFT):
            self._sizer.Add(spacer, 0)
            self._sizer.Add(tool, 0, align|align2)
        else:
            self._sizer.Add(spacer, 0)
            self._sizer.Add(tool, 0, align|align2)

    def GetControlSizer(self):
        """Get the sizer that is used to layout the contols (horizontal sizer)
        @return: wx.BoxSizer

        """
        return self._sizer

    def GetControlSpacing(self):
        """Get the spacing used between controls
        @return: size tuple

        """
        return self._spacing

    def IsVerticalMode(self):
        """Is the ControlBar in vertical orientation
        @return: bool

        """
        return self._style & CTRLBAR_STYLE_VERTICAL

    def DoPaintBackground(self, dc, rect, color, color2):
        """Paint the background of the given rect based on the style of
        the control bar.
        @param dc: DC to draw on
        @param rect: wx.Rect
        @param color: Pen/Base gradient color
        @param color2: Gradient end color

        """
        # Paint the gradient
        if self._style & CTRLBAR_STYLE_GRADIENT:
            if isinstance(dc, wx.GCDC):
                gc = dc.GetGraphicsContext()
            else:
                gc = wx.GraphicsContext.Create(dc)

            if gc is None:
                return

            if not self.IsVerticalMode():
                grad = gc.CreateLinearGradientBrush(rect.x, rect.y, rect.x,
                                                    rect.x+rect.height,
                                                    color2, color)
            else:
                grad = gc.CreateLinearGradientBrush(rect.x, rect.y,
                                                    rect.x+rect.width,
                                                    rect.y,
                                                    color2, color)

            gc.SetPen(gc.CreatePen(self._pen))
            gc.SetBrush(grad)
            gc.DrawRectangle(rect.x, rect.y, rect.Width - 0.5, rect.Height - 0.5)

        dc.SetPen(wx.Pen(color, 1))

        # TODO: handle vertical mode
        if not self.IsVerticalMode():
            # Add a border to the bottom
            if self._style & CTRLBAR_STYLE_BORDER_BOTTOM:
                dc.DrawLine(rect.x, rect.GetHeight() - 1,
                            rect.GetWidth(), rect.GetHeight() - 1)

            # Add a border to the top
            if self._style & CTRLBAR_STYLE_BORDER_TOP:
                dc.DrawLine(rect.x, 1, rect.GetWidth(), 1)

    def OnPaint(self, evt):
        """Paint the background to match the current style
        @param evt: wx.PaintEvent

        """
        dc = wx.AutoBufferedPaintDCFactory(self)
        gc = wx.GCDC(dc)
        rect = self.GetClientRect()

        self.DoPaintBackground(gc, rect, self._color, self._color2)

        evt.Skip()

    def SetToolSpacing(self, px):
        """Set the spacing to use between tools/controls.
        @param px: int (number of pixels)
        @todo: dynamically update existing layouts

        """
        self._spacing = (px, px)

    def SetVMargin(self, top, bottom):
        """WARNING this method is Deprecated use SetMargins instead!!
        @param top: Top margin in pixels
        @param bottom: Bottom margin in pixels

        """
        # TODO: Remove all usage of this method
        self.SetMargins(top, bottom)

    def SetMargins(self, param1, param2):
        """Setup the margins on the edges of the ControlBar
        @param param1: left/top margin depending on orientation
        @param param2: right/bottom margin depending on orientation

        """
        sizer = self.GetSizer()
        if wx.VERSION < (2, 9, 0, 0, ''):
            sizer.GetItem(0).SetSpacer((param1, param1))
            sizer.GetItem(2).SetSpacer((param2, param2))
        else:
            sizer.GetItem(0).AssignSpacer((param1, param1))
            sizer.GetItem(2).AssignSpacer((param2, param2))
        sizer.Layout()

    def SetWindowStyle(self, style):
        """Set the style flags of this window
        @param style: long

        """
        if self.IsVerticalMode() and not (CTRLBAR_STYLE_VERTICAL & style):
            # Switching from vertical to HORIZONTAL
            self._sizer.SetOrientation(wx.HORIZONTAL)
        elif not self.IsVerticalMode() and (CTRLBAR_STYLE_VERTICAL & style):
            # Switching from horizontal to vertical
            self._sizer.SetOrientation(wx.VERTICAL)
        self._style = style
        self.Layout()
        self.Refresh()

#--------------------------------------------------------------------------#

class _SegmentButton(object):
    """Class for managing segment button data"""
    def __init__(self, id_, bmp, label, lbl_size):
        super(_SegmentButton, self).__init__()

        # Attributes
        self._id = id_
        self._bmp = bmp
        self._lbl = label
        self._lbl_size = lbl_size
        self._rect = wx.Rect()
        self._bx1 = 0
        self._bx2 = 0
        self._opts = 0
        self._selected = False
        self._x_button = wx.Rect()
        self._x_state = SEGMENT_STATE_NONE

    Id = property(lambda self: self._id,
                  lambda self, id_: setattr(self, '_id', id_))
    Bitmap = property(lambda self: self._bmp,
                      lambda self, bmp: setattr(self, '_bmp', bmp))
    Label = property(lambda self: self._lbl,
                     lambda self, label: setattr(self, '_lbl', label))
    LabelSize = property(lambda self: self._lbl_size,
                         lambda self, size: setattr(self, '_lbl_size', size))
    Rect = property(lambda self: self._rect,
                    lambda self, rect: setattr(self, '_rect', rect))
    Selected = property(lambda self: self._selected,
                        lambda self, sel: setattr(self, '_selected', sel))
    XState = property(lambda self: self._x_state,
                      lambda self, state: setattr(self, '_x_state', state))
    XButton = property(lambda self: self._x_button,
                       lambda self, rect: setattr(self, '_x_button', rect))
    BX1 = property(lambda self: self._bx1,
                   lambda self, x1: setattr(self, '_bx1', x1))
    BX2 = property(lambda self: self._bx2,
                   lambda self, x2: setattr(self, '_bx2', x2))
    Options = property(lambda self: self._opts,
                       lambda self, opt: setattr(self, '_opts', opt))

class SegmentBar(ControlBar):
    """Simple toolbar like control that displays bitmaps and optionally
    labels below each bitmap. The bitmaps are turned into a toggle button
    where only one segment in the bar can be selected at one time.

    """
    HPAD = 5
    VPAD = 3
    def __init__(self, parent, id=wx.ID_ANY,
                 pos=wx.DefaultPosition, size=wx.DefaultSize,
                 style=CTRLBAR_STYLE_DEFAULT,
                 name=SEGBAR_NAME_STR):
        super(SegmentBar, self).__init__(parent, id, pos, size, style, name)

        # Attributes
        self._buttons = list() # list of _SegmentButtons
        self._segsize = (0, 0)
        self._selected = -1
        self._scolor1 = AdjustColour(self._color, -20)
        self._scolor2 = AdjustColour(self._color2, -20)
        self._spen = wx.Pen(AdjustColour(self._pen.GetColour(), -25))
        self._x_clicked_before = False
        self._tip_timer = wx.Timer(self)

        if wx.Platform == '__WXMAC__':
            self.SetWindowVariant(wx.WINDOW_VARIANT_SMALL)

        # Event Handlers
        self.Bind(wx.EVT_ENTER_WINDOW, self.OnEnter)
        self.Bind(wx.EVT_ENTER_WINDOW, self.OnLeave)
        self.Bind(wx.EVT_LEFT_DOWN, self.OnLeftDown)
        self.Bind(wx.EVT_LEFT_UP, self.OnLeftUp)
        self.Bind(wx.EVT_MOTION, self.OnMouseMove)
        self.Bind(wx.EVT_TIMER, self.OnTipTimer, self._tip_timer)
        self.Bind(wx.EVT_ERASE_BACKGROUND, self.OnEraseBackground)
        self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy)

    def _RestartTimer(self):
        """Reset the tip timer for showing tooltips when
        the segments have their labels hidden.

        """
        if not (self._style & CTRLBAR_STYLE_LABELS):
            if self._tip_timer.IsRunning():
                self._tip_timer.Stop()
            self._tip_timer.Start(1000, True)

    def OnDestroy(self, evt):
        """Cleanup on Destroy"""
        if evt.GetEventObject() is self:
            if self._tip_timer.IsRunning():
                self._tip_timer.Stop()
        evt.Skip()

    def AddSegment(self, id, bmp, label=u''):
        """Add a segment to the bar
        @param id: button id
        @param bmp: wx.Bitmap
        @param label: string

        """
        assert bmp.IsOk()
        lsize = self.GetTextExtent(label)
        segment = _SegmentButton(id, bmp, label, lsize)
        self._buttons.append(segment)
        self.InvalidateBestSize()
        self.Refresh()

    def DoDrawButton(self, dc, pos, bidx, selected=False, draw_label=False):
        """Draw a button
        @param dc: DC to draw on
        @param pos: X coordinate (horizontal mode) / Y coordinate (vertical mode)
        @param bidx: button dict
        @keyword selected: is this the selected button (bool)
        @keyword draw_label: draw the label (bool)
        return: int (next xpos)

        """
        button = self._buttons[bidx]
        height = self.Rect.height
        bVertical = self.IsVerticalMode()

        # Draw the background of the button
        if not bVertical:
            rside = pos + self._segsize[0]
            brect = wx.Rect(pos, 0, rside - pos, height)
        else:
            rside = pos + self._segsize[1]
            brect = wx.Rect(0, pos, self.Rect.Width, rside - pos)

        button.Rect = brect
        if selected:
            self.DoPaintBackground(dc, brect, self._scolor1, self._scolor2)

        # Draw the bitmap
        bmp = button.Bitmap
        bsize = button.Bitmap.Size
        if not bVertical:
            bxpos = ((self._segsize[0] / 2) - (bsize.width / 2)) + pos
            bpos = (bxpos, SegmentBar.VPAD)
        else:
            bxpos = ((self._segsize[0] / 2) - (bsize.width / 2))
            bpos = (bxpos, pos + SegmentBar.VPAD)
        dc.DrawBitmap(bmp, bpos[0], bpos[1], bmp.Mask != None)

        if draw_label:
            lcolor = wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOWTEXT)
            dc.SetTextForeground(lcolor)
            twidth, theight = button.LabelSize
            if not bVertical:
                lxpos = pos
                typos = height - theight - 2
            else:
                lxpos = 0
                typos = pos + (self._segsize[1] - theight - 2)
            trect = wx.Rect(lxpos, typos, self._segsize[0], theight + 3)
            dc.DrawLabel(button.Label, trect, wx.ALIGN_CENTER)

        if not selected:
            if not (self._style & CTRLBAR_STYLE_NO_DIVIDERS):
                dc.SetPen(self._pen)
                if not bVertical:
                    dc.DrawLine(pos, 0, pos, height)
                    dc.DrawLine(rside, 0, rside, height)
                else:
                    dc.DrawLine(0, pos, self.Rect.Width, pos)
                    dc.DrawLine(0, rside, self.Rect.Width, rside)
        else:
            dc.SetPen(self._spen)
            tmpx = pos + 1
            trside = rside - 1
            if not bVertical:
                dc.DrawLine(tmpx, 0, tmpx, height)
                dc.DrawLine(trside, 0, trside, height)
            else:
                dc.DrawLine(0, tmpx, self.Rect.Width, tmpx)
                dc.DrawLine(0, trside, self.Rect.Width, trside)

            tpen = wx.Pen(self._spen.Colour)
            tpen.SetJoin(wx.JOIN_BEVEL)
            if not bVertical:
                mpoint = height / 2
                dc.DrawLine(tmpx + 1, mpoint, tmpx, 0)
                dc.DrawLine(tmpx + 1, mpoint, tmpx, height)
                dc.DrawLine(trside - 1, mpoint, trside, 0)
                dc.DrawLine(trside - 1, mpoint, trside, height)
            else:
                mpoint = self.Rect.Width / 2
                dc.DrawLine(mpoint, tmpx + 1, 0, tmpx)
                dc.DrawLine(mpoint, tmpx + 1, self.Rect.Width, tmpx)
                dc.DrawLine(mpoint, trside - 1, 0, trside)
                dc.DrawLine(mpoint, trside - 1, self.Rect.Width, trside)

        # Update derived button data
        button.BX1 = pos + 1
        button.BX2 = rside - 1
        button.Selected = selected

        # Draw delete button if button has one
        if self.SegmentHasCloseButton(bidx):
            if not bVertical:
                brect = wx.Rect(button.BX1, 0, button.BX2 - (pos - 1), height)
            else:
                brect = wx.Rect(0, button.BX1, self.Rect.Width, 
                                button.BX2 - (pos - 1))
            self.DoDrawCloseBtn(dc, button, brect)
        return rside # right/bottom of button just drawn

    def DoDrawCloseBtn(self, gcdc, button, rect):
        """Draw the close button on the segment
        @param gcdc: Device Context
        @param button: Segment Dict
        @param rect: Segment Rect

        """
        if button.Options & SEGBTN_OPT_CLOSEBTNL:
            x = rect.x + 8
            y = rect.y + 6
        else:
            x = (rect.x + rect.Width) - 8
            y = rect.y + 6

        color = self._scolor2
        if button.Selected:
            color = AdjustColour(color, -25)

        if button.XState == SEGMENT_STATE_X:
            color = AdjustColour(color, -20)

        gcdc.SetPen(wx.Pen(AdjustColour(color, -30)))

        brect = wx.Rect(x-3, y-3, 8, 8)
        bmp = DrawCircleCloseBmp(color, wx.WHITE)
        gcdc.DrawBitmap(bmp, brect.x, brect.y)
        button.XButton = brect
        return
    
        # Square style button
#        gcdc.DrawRectangleRect(brect)
#        gcdc.SetPen(wx.BLACK_PEN)
#        gcdc.DrawLine(brect.x+1, brect.y+1,
#                      brect.x + brect.GetWidth() - 1, brect.y + brect.GetHeight() - 1)
#        gcdc.DrawLine(brect.x + brect.GetWidth() - 1, brect.y + 1,
#                      brect.x + 1, brect.y + brect.GetHeight() - 1)
#        gcdc.SetBrush(brush)
#        gcdc.SetPen(pen)
#        button['xbtn'] = brect

    def DoGetBestSize(self):
        """Get the best size for the control"""
        mwidth, mheight = 0, 0
        draw_label = self._style & CTRLBAR_STYLE_LABELS
        for btn in self._buttons:
            bwidth, bheight = btn.Bitmap.Size
            twidth = btn.LabelSize[0]
            if bheight > mheight:
                mheight = bheight

            if bwidth > mwidth:
                mwidth = bwidth

            if draw_label:
                if twidth > mwidth:
                    mwidth = twidth

        # Adjust for label text
        if draw_label and len(self._buttons):
            mheight += self._buttons[0].LabelSize[1]

        if self.IsVerticalMode():
            height = (mheight + (SegmentBar.VPAD * 2) * len(self._buttons))
            width = mwidth
        else:
            width = (mwidth + (SegmentBar.HPAD * 2)) * len(self._buttons)
            height = mheight

        size = wx.Size(width + (SegmentBar.HPAD * 2),
                       height + (SegmentBar.VPAD * 2))
        self.CacheBestSize(size)
        self._segsize = (mwidth + (SegmentBar.HPAD * 2),
                         mheight + (SegmentBar.VPAD * 2))
        return size

    def GetIndexFromPosition(self, pos):
        """Get the segment index closest to the given position"""
        if not self.IsVerticalMode():
            cur_x = pos[0]
        else:
            cur_x = pos[1]
        for idx, button in enumerate(self._buttons):
            xpos = button.BX1
            xpos2 = button.BX2
            if cur_x >= xpos and cur_x <= xpos2 + 1:
                return idx
        else:
            return wx.NOT_FOUND

    def GetSegmentCount(self):
        """Get the number segments in the control
        @return: int

        """
        return len(self._buttons)

    def GetSegmentLabel(self, index):
        """Get the label of the given segment
        @param index: segment index
        @return: string

        """
        return self._buttons[index].Label

    def GetSelection(self):
        """Get the currently selected index"""
        return self._selected

    def HitTest(self, pos):
        """Find where the position is in the window
        @param pos: (x, y) in client cords
        @return: int

        """
        index = self.GetIndexFromPosition(pos)
        where = SEGMENT_HT_NOWHERE
        if index != wx.NOT_FOUND:
            button = self._buttons[index]
            if self.SegmentHasCloseButton(index):
                brect = button.XButton
                trect = wx.Rect(brect.x, brect.y, brect.Width+4, brect.Height+4)
                if trect.Contains(pos):
                    where = SEGMENT_HT_X_BTN
                else:
                    where = SEGMENT_HT_SEG
            else:
                where = SEGMENT_HT_SEG

        return where, index

    def OnEraseBackground(self, evt):
        """Handle the erase background event"""
        pass

    def OnLeftDown(self, evt):
        """Handle clicks on the bar
        @param evt: wx.MouseEvent

        """
        epos = evt.GetPosition()
        index = self.GetIndexFromPosition(epos)
        if index != wx.NOT_FOUND:
            button = self._buttons[index]
            pre = self._selected
            self._selected = index

            if self._selected != pre:
                self.Refresh()
                sevt = SegmentBarEvent(edEVT_SEGMENT_SELECTED, button.Id)
                sevt.SetSelections(pre, index)
                sevt.SetEventObject(self)
                self.GetEventHandler().ProcessEvent(sevt)

        self._x_clicked_before = False

        # Check for click on close btn
        if self.SegmentHasCloseButton(index):
            if self.HitTest(epos)[0] == SEGMENT_HT_X_BTN:
                self._x_clicked_before = True

        evt.Skip()

    def OnLeftUp(self, evt):
        """Handle clicks on the bar
        @param evt: wx.MouseEvent

        """
        epos = evt.GetPosition()
        where, index = self.HitTest(epos)

        # Check for click on close btn
        if self.SegmentHasCloseButton(index) and self._x_clicked_before:
            if where == SEGMENT_HT_X_BTN:
                event = SegmentBarEvent(edEVT_SEGMENT_CLOSE, self.GetId())
                event.SetSelections(index, index)
                event.SetEventObject(self)
                self.GetEventHandler().ProcessEvent(event)
                if not event.IsAllowed():
                    return False
                removed = self.RemoveSegment(index)

        evt.Skip()

    def OnEnter(self, evt):
        """Mouse has entered the SegmentBar, update state info"""
        evt.Skip()

    def OnLeave(self, evt):
        """Mouse has left the SegmentBar, update state info"""
        if self._tip_timer.IsRunning():
            self._tip_timer.Stop()
        evt.Skip()

    def OnMouseMove(self, evt):
        """Handle when the mouse moves over the bar"""
        epos = evt.GetPosition()
        where, index = self.HitTest(epos)
        if index == -1:
            return
        if not self.SegmentHasCloseButton(index):
            self._RestartTimer()
            return

        # Update button state
        button = self._buttons[index]
        x_state = button.XState
        button.XState = SEGMENT_STATE_NONE

        if where != SEGMENT_HT_NOWHERE:
            if where == SEGMENT_HT_X_BTN:
                button.XState = SEGMENT_STATE_X
            elif where == SEGMENT_HT_SEG:
                # TODO: add highlight option for hover on segment
                pass
        else:
            self._RestartTimer()
            evt.Skip()
            return

        # If the hover state over a segments close button
        # has changed redraw the close button to reflect the
        # proper state.
        if button.XState != x_state:
            crect = self.GetClientRect()
            if not self.IsVerticalMode():
                brect = wx.Rect(button.BX1, 0,
                                button.BX2 - (button.BX1 - 2),
                                crect.Height)
            else:
                brect = wx.Rect(button.BX1, 0,
                                crect.Width,
                                button.BX2 - (button.BX1 - 2))
            self.Refresh(False, brect)
        self._RestartTimer()
        evt.Skip()

    def OnTipTimer(self, evt):
        """Show the tooltip for the current SegmentButton"""
        pos = self.ScreenToClient(wx.GetMousePosition())
        where, index = self.HitTest(pos)
        if index != -1:
            button = self._buttons[index]
            if button.Label:
                rect = button.Rect
                x,y = self.ClientToScreenXY(rect.x, rect.y)
                rect.x = x
                rect.y = y
                wx.TipWindow(self, button.Label, rectBound=rect) # Transient

    def OnPaint(self, evt):
        """Paint the control"""
        dc = wx.AutoBufferedPaintDCFactory(self)
        gc = wx.GCDC(dc)

        # Setup
        dc.SetBrush(wx.TRANSPARENT_BRUSH)
        gc.SetBrush(wx.TRANSPARENT_BRUSH)
        gc.SetFont(self.GetFont())
        gc.SetBackgroundMode(wx.TRANSPARENT)
        gc.Clear()

        # Paint the background
        rect = self.GetClientRect()
        self.DoPaintBackground(gc, rect, self._color, self._color2)

        # Draw the buttons
        # TODO: would be more efficient to just redraw the buttons that
        #       need redrawing.
        npos = SegmentBar.HPAD
        if self.IsVerticalMode():
            npos = SegmentBar.VPAD
        use_labels = self._style & CTRLBAR_STYLE_LABELS
        for idx, button in enumerate(self._buttons):
            npos = self.DoDrawButton(gc, npos, idx,
                                     self._selected == idx,
                                     use_labels)

    def RemoveSegment(self, index):
        """Remove a segment from the bar
        @param index: int
        @return: bool

        """
        button = self._buttons[index]

        # TODO: wxPython 2.8.9.2 this causes a crash...
#        if button['bmp']:
#            button['bmp'].Destroy()
        del self._buttons[index]

        if self.GetSelection() == index:
            count = self.GetSegmentCount()
            if index >= count:
                self.SetSelection(count-1)

        self.Refresh()
        return True

    def SegmentHasCloseButton(self, index):
        """Does the segment at index have a close button
        @param index: int

        """
        button = self._buttons[index]
        if button.Options & SEGBTN_OPT_CLOSEBTNL or \
           button.Options & SEGBTN_OPT_CLOSEBTNR:
            return True
        return False

    def SetSegmentImage(self, index, bmp):
        """Set the image to use on the given segment
        @param index: int
        @param bmp: Bitmap

        """
        assert bmp.IsOk()
        segment = self._buttons[index]
        if segment.Bitmap.IsOk():
            segment.Bitmap.Destroy()
            segment.Bitmap = None
        segment.Bitmap = bmp
        self.InvalidateBestSize()
        self.Refresh()

    def SetSegmentLabel(self, index, label):
        """Set the label for a given segment
        @param index: segment index
        @param label: string

        """
        segment = self._buttons[index]
        lsize = self.GetTextExtent(label)
        segment.Label = label
        segment.LabelSize = lsize
        self.InvalidateBestSize()
        self.Refresh()

    def SetSegmentOption(self, index, option):
        """Set an option on a given segment
        @param index: segment index
        @param option: option to set

        """
        button = self._buttons[index]
        button.Options = button.Options|option
        self.Refresh()

    def SetSelection(self, index):
        """Set the selection
        @param index: int

        """
        self._selected = index
        self.Refresh()

# Cleanup namespace
#del SegmentBar.__dict__['AddControl']
#del SegmentBar.__dict__['AddSpacer']
#del SegmentBar.__dict__['AddTool']
#del SegmentBar.__dict__['SetToolSpacing']
#del SegmentBar.__dict__['SetVMargin']

#--------------------------------------------------------------------------#



###############################################################################
# Name: ed_shelf.py                                                           #
# Purpose: Editra Shelf container                                             #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Shelf plugin and control implementation

@summary: Shelf Implementation

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: ed_shelf.py 72821 2012-10-29 22:48:49Z CJP $"
__revision__ = "$Revision: 72821 $"

#-----------------------------------------------------------------------------#
# Imports
import re
import wx

# Editra Imports
import ed_menu
import ed_glob
from profiler import Profile_Get
import ed_msg
import plugin
import iface
import ed_fmgr
from extern import aui
import ed_book
import ebmlib

#--------------------------------------------------------------------------#
# Globals
PGNUM_PAT = re.compile(' - [0-9]+')
_ = wx.GetTranslation

#-----------------------------------------------------------------------------#

class Shelf(plugin.Plugin):
    """Plugin that creates a notebook for holding the various Shelf items
    implemented by L{ShelfI}.

    """
    SHELF_NAME = u'Shelf'
    observers = plugin.ExtensionPoint(iface.ShelfI)

    def GetUiHandlers(self, delegate):
        """Gets the update ui handlers for the shelf's menu
        @param delegate: L{EdShelfDelegate} instance
        @return: [(ID, handler),]

        """
        handlers = [ (item.GetId(), delegate.UpdateShelfMenuUI)
                     for item in self.observers ]
        return handlers

    def Init(self, parent):
        """Mixes the shelf into the parent window
        @param parent: Reference to MainWindow
        @return: L{EdShelfDelegate}

        """
        # First check if the parent has an instance already
        mgr = parent.GetFrameManager()
        if mgr.GetPane(Shelf.SHELF_NAME).IsOk():
            return

        # HACK - fixes mouse event issues that result in wrong
        #        tab indexes being reported in the notebook.
        wrapper = ShelfWrapper(parent)
        shelf = wrapper.GetShelf()
        mgr.AddPane(wrapper,
                    ed_fmgr.EdPaneInfo().Name(Shelf.SHELF_NAME).\
                            Caption(_("Shelf")).Bottom().Layer(0).\
                            CloseButton(True).MaximizeButton(True).\
                            BestSize(wx.Size(500,250)))

        # Hide the pane and let the perspective manager take care of it
        mgr.GetPane(Shelf.SHELF_NAME).Hide()
        mgr.Update()

        # Create the delegate
        delegate = EdShelfDelegate(shelf, self)

        # Install Shelf menu under View and bind event handlers
        view = parent.GetMenuBar().GetMenuByName("view")
        menu = delegate.GetMenu()
        pos = 0
        for pos in range(view.GetMenuItemCount()):
            mitem = view.FindItemByPosition(pos)
            if mitem.GetId() == ed_glob.ID_PERSPECTIVES:
                break

        view.InsertMenu(pos + 1, ed_glob.ID_SHELF, _("Shelf"), 
                        menu, _("Put an item on the Shelf"))

        for item in menu.GetMenuItems():
            if item.IsSeparator():
                continue
            parent.Bind(wx.EVT_MENU, delegate.OnGetShelfItem, item)

        if menu.GetMenuItemCount() < 3:
            view.Enable(ed_glob.ID_SHELF, False)

        # Check for any other plugin specific install needs
        for observer in self.observers:
            if not observer.IsInstalled() and \
               hasattr(observer, 'InstallComponents'):
                observer.InstallComponents(parent)

        # Only Load Perspective if all items are loaded
        if delegate.StockShelf(Profile_Get('SHELF_ITEMS', 'list', [])):
            delegate.SetPerspective(Profile_Get('SHELF_LAYOUT', 'str', u""))
            delegate.SetSelection(Profile_Get('SHELF_SELECTION', 'int', -1))
        return delegate

#--------------------------------------------------------------------------#

class EdShelfBook(ed_book.EdBaseBook):
    ID_SHELF_SUBMENU = wx.NewId()
    ID_CLOSE_LIKE_TABS = wx.NewId()
    """Shelf notebook control"""
    def __init__(self, parent):
        style = aui.AUI_NB_BOTTOM | \
                aui.AUI_NB_TAB_SPLIT | \
                aui.AUI_NB_SCROLL_BUTTONS | \
                aui.AUI_NB_CLOSE_ON_ACTIVE_TAB | \
                aui.AUI_NB_TAB_MOVE | \
                aui.AUI_NB_DRAW_DND_TAB
        super(EdShelfBook, self).__init__(parent, style=style)

        # Attributes
        self._open = dict()
        self._name2idx = dict() # For settings maintenance
        self._menu = ebmlib.ContextMenuManager()
        self._mcback = None
        self._mw = None

        # Event Handlers
        # HACK: binding to self works differently than parent catching it
        #       binding here causes incorrect mouse coords to be sent to
        #       event handler during click events!
#        self.Bind(aui.EVT_AUINOTEBOOK_TAB_RIGHT_UP, self.OnTabRightUp)
        self.Bind(aui.EVT_AUINOTEBOOK_BG_RIGHT_UP, self.OnBgRightUp)
        self.Bind(aui.EVT_AUINOTEBOOK_PAGE_CLOSE, self.OnItemClose)
        self.Bind(wx.EVT_MENU, self.OnTabMenu)
        self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy, self)

        # Message handlers
        ed_msg.Subscribe(self.OnUpdateTabs, ed_msg.EDMSG_THEME_NOTEBOOK)
        ed_msg.Subscribe(self.OnUpdateTabs, ed_msg.EDMSG_THEME_CHANGED)

    def OnDestroy(self, evt):
        if evt.GetId() == self.GetId():
            self._menu.Clear()
            ed_msg.Unsubscribe(self.OnUpdateTabs)
        evt.Skip()

    def OnTabRightUp(self, evt):
        """Tab right click handler"""
        self._menu.Clear()
        if self.MenuCallback:
            sel = self.GetPageIndex(evt.Page)
            s_lbl = self.GetPageText(sel)
            s_lbl = s_lbl.rsplit('-')[0].strip()
            tab_menu = ed_menu.EdMenu()
            tab_menu.Append(EdShelfBook.ID_CLOSE_LIKE_TABS,
                            _("Close All '%s'") % s_lbl)
            self._menu.AddHandler(EdShelfBook.ID_CLOSE_LIKE_TABS,
                                  lambda tab, evt: self.CloseAll(s_lbl))
            tab_menu.AppendSeparator()
            shelf_menu = self.MenuCallback()
            tab_menu.AppendMenu(EdShelfBook.ID_SHELF_SUBMENU, 
                                _("Open"), shelf_menu)
            self._menu.Menu = tab_menu
            self.PopupMenu(self._menu.Menu)

    def OnBgRightUp(self, evt):
        """Show context menu"""
        self._menu.Clear()
        if self.MenuCallback:
            self._menu.Menu = self.MenuCallback()
            self.PopupMenu(self._menu.Menu)

    def OnTabMenu(self, evt):
        """Handle tab menu events"""
        handler = self._menu.GetHandler(evt.Id)
        if handler is not None:
            handler(self.GetCurrentPage(), evt)
        else:
            evt.Skip()

    MainWindow = property(lambda self: self._mw,
                          lambda self, mw: setattr(self, '_mw', mw))
    BitmapCallbacks = property(lambda self: self._name2idx)
    MenuCallback = property(lambda self: self._mcback,
                            lambda self, funct: setattr(self, '_mcback', funct))

    def AddItem(self, item, name, bmp=wx.NullBitmap):
        """Add an item to the shelf's notebook. This is useful for interacting
        with the Shelf from outside its interface. It may be necessary to
        call L{EnsureShelfVisible} before or after adding an item if you wish
        the shelf to be shown when the item is added.
        @param item: A panel like instance to add to the shelf's notebook
        @param name: Items name used for page text in notebook
        @keyword bmp: Tab bitmap to display

        """
        self.AddPage(item, 
                     u"%s - %d" % (name, self._open.get(name, 0)),
                     select=True)

        # Set the tab icon
        self.SetPageBitmap(self.GetPageCount()-1, bmp)
        self._open[name] = self._open.get(name, 0) + 1

    def OnItemClose(self, evt):
        """Callback for when pages are closed"""
#        name = self.GetPageText(evt.GetSelection())
#        name = name.rsplit('-')[0].strip()
#        self._open[name] = self._open.get(name, 1) - 1
        evt.Skip()

    def CloseAll(self, name):
        """Close all of 'name' shelf items
        @param name: shelf item name (string)

        """
        idxs = self.GetAllIndexes(name)
        idxs.reverse()
        for idx in idxs:
            self.DeletePage(idx)

    def GetAllIndexes(self, name):
        """Get all the indexes of pages of the item identified
        by 'name'.
        @param name: shelf item name (string)

        """
        idxs = list()
        for pg in range(self.GetPageCount()):
            lbl = self.GetPageText(pg)
            lbl = PGNUM_PAT.sub(u"", lbl)
            if lbl == name:
                idxs.append(pg)
        return idxs

    def EnsureShelfVisible(self):
        """Make sure the Shelf is visible
        @precondition: Shelf.Init has been called
        @postcondition: Shelf is shown

        """
        mgr = self.MainWindow.GetFrameManager()
        pane = mgr.GetPane(Shelf.SHELF_NAME)
        if not pane.IsShown():
            pane.Show()
            mgr.Update()

    def GetCount(self, item_name):
        """Get the number of open instances of a given Shelf Item
        @param item_name: Name of the Shelf item
        @return: number of instances on the Shelf

        """
        count = 0
        for page in range(self.GetPageCount()):
            if self.GetPageText(page).startswith(item_name):
                count = count + 1
        return count

    def GetMainWindow(self):
        """Get the main window that this shelf instance was created for
        @return: ed_main.MainWindow

        """
        return self.MainWindow

    def GetOpen(self):
        """Get the list of open shelf items
        @return: list

        """
        return self._open

    def Hide(self):
        """Hide the shelf
        @postcondition: Shelf is hidden by aui manager

        """
        mgr = self.MainWindow.GetFrameManager()
        pane = mgr.GetPane(Shelf.SHELF_NAME)
        if pane.IsOk():
            pane.Hide()
            mgr.Update()

    def ItemIsOnShelf(self, item_name):
        """Check if at least one instance of a given item
        is currently on the Shelf.
        @param item_name: name of Item to look for

        """
        for page in range(self.GetPageCount()):
            if self.GetPageText(page).startswith(item_name):
                return True
        return False

    def IsShown(self):
        """Is the shelf visible?
        @return: bool

        """
        mgr = self.MainWindow.GetFrameManager()
        pane = mgr.GetPane(Shelf.SHELF_NAME)
        if pane.IsOk():
            return pane.IsShown()
        else:
            return False

    def OnUpdateTabs(self, msg):
        """Update all tab images depending upon current settings"""
        if not self:
            return
        
        if not Profile_Get('TABICONS', default=True):
            for page in range(self.GetPageCount()):
                self.SetPageBitmap(page, wx.NullBitmap)
        else:
            # Show the icons
            for pnum in range(self.GetPageCount()):
                page = self.GetPage(pnum)
                bmp = self.BitmapCallbacks.get(repr(page.__class__), lambda:wx.NullBitmap)()
                self.SetPageBitmap(pnum, bmp)

#--------------------------------------------------------------------------#

class EdShelfDelegate(object):
    """Delegate class to mediate between the plugin singleton object and the
    UI implementation.

    """
    def __init__(self, shelf, pobject):
        """Create the delegate object
        @param shelf: Ui component instance
        @param pobject: Reference to the plugin object

        """
        super(EdShelfDelegate, self).__init__()

        # Attributes
        self._log = wx.GetApp().GetLog()
        self._shelf = shelf
        self._pin = pobject

        # Setup
        self._shelf.MenuCallback = getattr(self, 'GetShelfObjectMenu')

    @property
    def observers(self):
        return self._pin.observers

    def AddItem(self, item, name, bmp=wx.NullBitmap):
        """Add an item to the shelf"""
        self._shelf.AddItem(item, name, bmp)

    def CanStockItem(self, item_name):
        """See if a named item can be stocked or not, meaning if it
        can be saved and opened in the next session or not.
        @param item_name: name of item to check
        @return: bool whether item can be stocked or not

        """
        for item in self.observers:
            if item_name == item.GetName():
                if hasattr(item, 'IsStockable'):
                    return item.IsStockable()
                else:
                    break
        return False

    def EnsureShelfVisible(self):
        """Ensure the shelf is visible"""
        self._shelf.EnsureShelfVisible()

    def GetItemById(self, itemid):
        """Get the shelf item by its id
        @param itemid: Shelf item id
        @return: reference to a ShelfI object

        """
        for item in self.observers:
            if item.GetId() == itemid:
                return item
        return None

    def GetItemId(self, item_name):
        """Get the id that identifies a given item
        @param item_name: name of item to get ID for
        @return: integer id or None if not found

        """
        for item in self.observers:
            if item_name == item.GetName():
                return item.GetId()
        return None

    def GetItemStack(self):
        """Returns a list of ordered named items that are open in the shelf
        @return: list of strings

        """
        rval = list()
        if self._shelf is not None:
            for page in range(self._shelf.GetPageCount()):
                rval.append(re.sub(PGNUM_PAT, u'', 
                            self._shelf.GetPageText(page), 1))
        return rval

    def GetSelection(self):
        """Get the index of the currently selected tab"""
        if self._shelf:
            return self._shelf.GetSelection()
        return -1

    def GetPerspective(self):
        """Get the auinotebook perspective data
        @return: string

        """
        return self._shelf.SavePerspective()

    def SetPerspective(self, pdata):
        """Set the aui notebooks perspective and layout
        @param pdata: perspective data string

        """
        if pdata:
            try:
                self._shelf.LoadPerspective(pdata)
                self._shelf.Update()
            except Exception, msg:
                self._log("[shelf][err] Failed LoadPerspective: %s" % msg)

    def GetShelfObjectMenu(self):
        """Get the minimal menu that lists all Shelf objects
        without the 'Show Shelf' item.
        @return: ed_menu.EdMenu

        """
        menu = ed_menu.EdMenu()
        menu_items = list()
        open_items = self._shelf.GetOpen()
        for observer in self.observers:
            # Register Observers
            open_items[observer.GetName()] = 0
            try:
                menu_i = observer.GetMenuEntry(menu)
                if menu_i is not None:
                    menu_items.append((menu_i.GetItemLabel(), menu_i))
            except Exception, msg:
                self._log("[shelf][err] %s" % str(msg))
        menu_items.sort()

        combo = 0
        for item in menu_items:
            combo += 1
            shortcut = u""
            if combo < 10:
                shortcut = u"\tCtrl+Alt+" + unicode(combo)
            nitem = menu.Append(item[1].Id, item[1].GetText() + shortcut)
            if item[1].Bitmap.IsOk():
                nitem.SetBitmap(item[1].Bitmap)
            item[1].Destroy()
        return menu

    def GetMenu(self):
        """Return the menu of this object
        @return: ed_menu.EdMenu()

        """
        menu = self.GetShelfObjectMenu()
        menu.Insert(0, wx.ID_SEPARATOR)
        menu.Insert(0, ed_glob.ID_SHOW_SHELF, _("Show Shelf") + \
                    ed_menu.EdMenuBar.keybinder.GetBinding(ed_glob.ID_SHOW_SHELF), 
                    _("Show the Shelf"))
        return menu

    def GetOwnerWindow(self):
        """Return the L{ed_main.MainWindow} instance that owns/created
        this Shelf.
        @return: reference to ed_main.MainWindow or None

        """
        return self._shelf.GetMainWindow()

    def GetWindow(self):
        """Return reference to the Shelfs window component
        @return: AuiNotebook

        """
        return self._shelf

    def OnGetShelfItem(self, evt):
        """Handles menu events that have been registered
        by the Shelf Items on the Shelf.
        @param evt: Event that called this handler

        """
        e_id = evt.GetId()
        if e_id == ed_glob.ID_SHOW_SHELF:
            parent = self.GetOwnerWindow()
            if self._shelf.IsShown():
                self._shelf.Hide()
                nb = parent.GetNotebook()
                nb.GetCurrentCtrl().SetFocus()
            else:
                self._shelf.EnsureShelfVisible()
                mgr = parent.GetFrameManager()
                pane = mgr.GetPane(Shelf.SHELF_NAME)
                if pane is not None:
                    page = pane.window.Book.GetCurrentPage()
                    if hasattr(page, 'SetFocus'):
                        page.SetFocus()
        else:
            self.PutItemOnShelf(evt.GetId())

    def OnPutShelfItemAway(self, evt):
        """Handles when an item is closed
        @param evt: event that called this handler
        @todo: is this needed?

        """
        raise NotImplementedError

    def PutItemOnShelf(self, shelfid):
        """Put an item on the shelf by using its unique shelf id.
        This is only for use with loading items implementing the
        L{ShelfI} interface. See L{AddItem} if you wish to pass
        a panel to the shelf to add.
        @param shelfid: id of the ShelfItem to open

        """
        item = None
        for shelfi in self.observers:
            if shelfi.GetId() == shelfid:
                item = shelfi
                break

        if item is None:
            return

        name = item.GetName()
        if self._shelf.ItemIsOnShelf(name) and \
            not item.AllowMultiple() or self._shelf is None:
            return
        else:
            self.EnsureShelfVisible()
            # Guard against crashes in creating plugin derived objects
            # log error to log and continue running.
            window = None
            try:
                window = item.CreateItem(self._shelf)
            except Exception, msg:
                self._log("[shelf][err] CreateItem failed: %s" % msg)
                return

            bmp = wx.NullBitmap
            if hasattr(item, 'GetBitmap'):
                self._shelf.BitmapCallbacks[repr(window.__class__)] = item.GetBitmap
                bmp = item.GetBitmap()
            else:
                self._shelf.BitmapCallbacks[repr(window.__class__)] = lambda:wx.NullBitmap
            self.AddItem(window, name, bmp)

    def RaiseItem(self, item_name):
        """Set the selection in the notebook to be the that of the first
        instance of item_name that is found in the shelf.
        @param item_name: ShelfI name
        @return: reference to the selected page or None if no instance is

        """
        for page in range(self._shelf.GetPageCount()):
            if self._shelf.GetPageText(page).startswith(item_name):
                self._shelf.SetSelection(page)
                return self._shelf.GetPage(page)
        else:
            return None

    def RaiseWindow(self, window):
        """Set the selection in the notebook to be the that of the given
        window. Mostly used internally by items implementing L{ShelfI}.
        @param window: Window object
        @return: reference to the selected page or None if no instance is

        """
        for page in range(self._shelf.GetPageCount()):
            ctrl = self._shelf.GetPage(page)
            if window == ctrl:
                self._shelf.SetSelection(page)
                return ctrl
        else:
            return None

    def SetSelection(self, index):
        """Select an item in the Shelf window
        @param index: shelf tab index

        """
        if self._shelf and index > 0 and index < self._shelf.GetPageCount():
            try:
                self._shelf.SetSelection(index)
            except Exception, msg:
                self._log("[shelf][err] Failed SetSelection: %s" % msg)

    def StockShelf(self, i_list):
        """Fill the shelf by opening an ordered list of items
        @param i_list: List of named L{ShelfI} instances
        @return: bool (True if all loaded / False otherwise)

        """
        bLoaded = True
        for item in i_list:
            if self.CanStockItem(item):
                itemid = self.GetItemId(item)
                if itemid:
                    self.PutItemOnShelf(itemid)
                else:
                    bLoaded = False
            else:
                bLoaded = False
        return bLoaded

    def UpdateShelfMenuUI(self, evt):
        """Enable/Disable shelf items based on whether they support
        muliple instances or not.
        @param evt: wxEVT_UPDATEUI

        """
        item = self.GetItemById(evt.GetId())
        if item is None:
            evt.Skip()
            return

        count = self._shelf.GetCount(item.GetName())
        if count and not item.AllowMultiple():
            evt.Enable(False)
        else:
            evt.Enable(True)

#-----------------------------------------------------------------------------#

class ShelfWrapper(wx.Panel):
    def __init__(self, parent):
        super(ShelfWrapper, self).__init__(parent)

        self._shelf = EdShelfBook(self)
        self._shelf.MainWindow = parent

        # Setup
        sizer = wx.BoxSizer()
        sizer.Add(self._shelf, 1, wx.EXPAND)
        self.SetSizer(sizer)

        self.Bind(aui.EVT_AUINOTEBOOK_TAB_RIGHT_UP, self.OnTabRightUp)

    Book = property(lambda self: self._shelf)

    def GetShelf(self):
        return self._shelf

    def OnTabRightUp(self, evt):
        self._shelf.OnTabRightUp(evt)



###############################################################################
# Name: filterdlg.py                                                          #
# Purpose: Filter dialog                                                      #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
Editra Control Library: FilterDialog


"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: filterdlg.py 65202 2010-08-06 15:49:23Z CJP $"
__revision__ = "$Revision: 65202 $"

__all__ = ["FilterDialog",]

#-----------------------------------------------------------------------------#
# Imports
import wx

# Eclib Imports
import ecbasewin

#-----------------------------------------------------------------------------#
# Globals

_ = wx.GetTranslation

#-----------------------------------------------------------------------------#

class FilterDialog(ecbasewin.ECBaseDlg):
    """Dialog that allows adding and removing items from a filter list"""
    def __init__(self, parent, id=wx.ID_ANY, title=u"",
                 pos=wx.DefaultPosition, size=wx.DefaultSize, 
                 style=wx.DEFAULT_DIALOG_STYLE, name=u"FilterDialog"):
        super(FilterDialog, self).__init__(parent, id, title,
                                           pos, size, style, name)

        # Attributes
        self.SetPanel(FilterPanel(self))

        # Event Handlers

#-----------------------------------------------------------------------------#

class FilterPanel(wx.Panel):
    """Filter dialog panel"""
    def __init__(self, parent):
        super(FilterPanel, self).__init__(parent)

        # Attributes
        self._left  = wx.ListBox(self, style=wx.LB_EXTENDED|wx.LB_SORT)
        self._right = wx.ListBox(self, style=wx.LB_EXTENDED)

        self.__DoLayout()

        # Event Handlers
        self.Bind(wx.EVT_BUTTON, self.OnButton)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateButton, id=wx.ID_ADD)
        self.Bind(wx.EVT_UPDATE_UI, self.OnUpdateButton, id=wx.ID_REMOVE)
        
    def __DoLayout(self):
        """Layout the panel"""
        vsizer = wx.BoxSizer(wx.VERTICAL)
        hsizer = wx.BoxSizer(wx.HORIZONTAL)

        hsizer.Add(self._left, 1, wx.EXPAND|wx.ALL, 10)

        # Add buttons
        bvsizer = wx.BoxSizer(wx.VERTICAL)
        addb = wx.Button(self, wx.ID_ADD, label=_("Add >>"))
        removeb = wx.Button(self, wx.ID_REMOVE, label=_("<< Remove"))
        bvsizer.AddStretchSpacer()
        bvsizer.AddMany([(addb, 0, wx.EXPAND),
                         ((10, 15), 0),
                         (removeb, 0, wx.EXPAND)])
        bvsizer.AddStretchSpacer()

        hsizer.Add(bvsizer, 0, wx.ALIGN_CENTER)
        hsizer.Add(self._right, 1, wx.EXPAND|wx.ALL, 10)

        vsizer.Add(hsizer, 1, wx.EXPAND)

        # Add main dialog buttons
        bsizer = wx.StdDialogButtonSizer()
        bsizer.AddButton(wx.Button(self, wx.ID_OK))
        btn = wx.Button(self, wx.ID_CANCEL)
        bsizer.AddButton(btn)
        btn.SetDefault()
        bsizer.Realize()
        vsizer.Add(bsizer, 0, wx.ALIGN_RIGHT)
        vsizer.AddSpacer(8)

        self.SetSizer(vsizer)
        self.SetAutoLayout(True)

    @ecbasewin.expose(FilterDialog)
    def GetIncludes(self):
        """Get the items from the includes list
        @return: list of strings

        """
        return self._right.GetItems()

    @ecbasewin.expose(FilterDialog)
    def SetIncludes(self, items):
        """Set the items in the includes list
        @param items: list of strings

        """
        return self._right.SetItems(items)

    @ecbasewin.expose(FilterDialog)
    def GetExcludes(self):
        """Get the items from the excludes list
        @return: list of strings

        """
        return self._left.GetItems()

    @ecbasewin.expose(FilterDialog)
    def SetExcludes(self, items):
        """set the items in the excludes list
        @param items: list of strings

        """
        return self._left.SetItems(items)

    @ecbasewin.expose(FilterDialog)
    def SetListValues(self, valuemap):
        """Set the values of the filter lists
        @param valuemap: dict(item=bool)

        """
        includes = list()
        excludes = list()
        for item, include in valuemap.iteritems():
            if include:
                includes.append(item)
            else:
                excludes.append(item)
        includes.sort()
        excludes.sort()
        self.SetIncludes(includes)
        self.SetExcludes(excludes)

    def OnButton(self, evt):
        e_id = evt.GetId()
        if e_id in (wx.ID_ADD, wx.ID_REMOVE):
            cmap = { wx.ID_ADD : (self._left, self._right),
                     wx.ID_REMOVE : (self._right, self._left) }
            idxs = list()
            for sel in cmap[e_id][0].GetSelections():
                selstr = cmap[e_id][0].GetString(sel)
                cmap[e_id][1].Append(selstr)
                idxs.append(sel)
            idxs.sort()
            idxs.reverse()
            for idx in idxs:
                cmap[e_id][0].Delete(idx)
        else:
            evt.Skip()

    def OnUpdateButton(self, evt):
        """Enable/Disable the Add/Remove buttons based on
        selections in the list.

        """
        e_id = evt.GetId()
        if e_id == wx.ID_ADD:
            evt.Enable(len(self._left.GetSelections()))
        elif e_id == wx.ID_REMOVE:
            evt.Enable(len(self._right.GetSelections()))
        else:
            evt.Skip()



###############################################################################
# Name: php.py                                                                #
# Purpose: Define PHP syntax for highlighting and other features              #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2007 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""
FILE: php.py
AUTHOR: Cody Precord
@summary: Lexer configuration module for PHP.

"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: _php.py 68798 2011-08-20 17:17:05Z CJP $"
__revision__ = "$Revision: 68798 $"

#-----------------------------------------------------------------------------#
# Imports
import wx.stc as stc

# Local Imports
import synglob
import syndata
import _html
from _cpp import AutoIndenter

#-----------------------------------------------------------------------------#

#---- Keyword Specifications ----#

# PHP Keywords
PHP_KEYWORDS = ("__LINE__ __FILE__ __FUNCTION__ __CLASS__ __METHOD__ declare "
                "else enddeclare endswitch elseif endif if switch as do endfor "
                "endforeach endwhile for foreach while case default switch "
                "break continue var bool boolean int integer real "
                "double float string array NULL extends global static "
                "new true false function "
                "class object self final public private protected try catch "
                "throw abstract parent interface implements "
                # Language Constructs
                "die echo empty exit eval include include_once isset list "
                "require require_once return print unset")

# PHP Standard Functions/Methods
# (roughly based off of PHP Pocket Reference by O'Reilly)
PHP_FUNC = ("__construct __autoload __destruct __get __set __isset __unset "
            "__call __sleep __wakeup __toString __set_state __clone "
            "apache_child_terminate apache_lookup_uri apache_note "
            "apache_request_headers apache_response_headers apache_setenv "
            "ascii2ebcdic ebcdic2ascii getallheaders virtual jewishtojd "
            "array_change_key_case array_chunk array_count_values "
            "array_diff_assoc array_diff array_fill array_filter array_flip "
            "array_intersect_assoc array_intersect array_key_exists array_keys "
            "array_map array_merge_recursive array_merge array_multisort "
            "array_pad array_pop array_push array_rand array_reduce array "
            "array_reverse array_search array_shift array_slice array_splice "
            "array_sum array_unique array_unshift array_values array_walk "
            "arsort asort compact count current each end extract in_array key "
            "krsort ksort natcasesort natsort next pos prev range reset "
            "rsort shuffle sizeof sort uasort uksort usort aspell_check "
            "aspell_new aspell_suggest bcadd bccomp bcdiv bcmod bcmul bcpow "
            "bcpowmod bcscale bcsqrt bcsub bzclose bzcompress bzdecompress "
            "bzerrno bzerror bzerrstr bzflush bzopen bzread bzwrite "
            "cal_days_in_month cal_from_jd cal_info cal_to_jd easter_date "
            "easter_days frenchtojd gregoriantojd jddayofweek jdmonthname "
            "jdtofrench jdtogregorian jdtojewish jdtojulian jdtounix "
            "juliantojd unixtojd ccvs_add ccvs_auth ccvs_command ccvs_count "
            "ccvs_delete ccvs_done ccvs_init ccvs_lookup ccvs_new ccvs_report "
            "ccvs_return ccvs_reverse ccvs_sale ccvs_status ccvs_textvalue "
            "ccvs_void call_user_method_array call_user_method class_exists "
            "get_class_methods get_class_vars get_class get_declared_classes "
            "get_object_vars get_parent_class is_a is_subclass_of com_load "
            "com_addref com_get com_invoke com_isenum com_load_typelib "
            "com_propget com_propput com_propset com_release com_set "
            "cpdf_add_annotation cpdf_add_outline cpdf_arc cpdf_begin_text "
            "cpdf_circle cpdf_clip cpdf_close cpdf_closepath_fill_stroke "
            "cpdf_closepath_stroke cpdf_closepath cpdf_continue_text "
            "cpdf_end_text cpdf_fill_stroke cpdf_fill cpdf_finalize_page "
            "cpdf_finalize cpdf_global_set_document_limits cpdf_import_jpeg "
            "cpdf_lineto cpdf_moveto cpdf_newpath cpdf_open cpdf_output_buffer "
            "cpdf_page_init cpdf_place_inline_image cpdf_rect cpdf_restore "
            "cpdf_rlineto cpdf_rmoveto cpdf_rotate_text cpdf_rotate "
            "cpdf_save_to_file cpdf_save cpdf_scale cpdf_set_action_url "
            "cpdf_set_char_spacing cpdf_set_creator cpdf_set_current_page "
            "cpdf_set_font_directories cpdf_set_font_map_file cpdf_set_font "
            "cpdf_set_horiz_scaling cpdf_set_keywords cpdf_set_leading "
            "cpdf_set_page_animation cpdf_set_subject cpdf_set_text_matrix "
            "cpdf_set_text_pos cpdf_set_text_rendering cpdf_set_text_rise "
            "cpdf_set_title cpdf_set_viewer_preferences cpdf_set_word_spacing "
            "cpdf_setdash cpdf_setflat cpdf_setgray_fill cpdf_setgray_stroke "
            "cpdf_setgray cpdf_setlinecap cpdf_setlinejoin cpdf_setlinewidth "
            "cpdf_setmiterlimit cpdf_setrgbcolor_fill cpdf_setrgbcolor_stroke "
            "cpdf_setrgbcolor cpdf_show_xy cpdf_show cpdf_stringwidth "
            "cpdf_text cpdf_translate crack_check crack_closedict cpdf_curveto "
            "crack_getlastmessage crack_opendict ctype_alnum ctype_alpha "
            "ctype_cntrl ctype_digit ctype_graph ctype_lower ctype_print "
            "ctype_punct ctype_space ctype_upper ctype_xdigit curl_close "
            "curl_errno curl_error curl_exec curl_getinfo curl_init chgrp "
            "curl_version checkdate date getdate gettimeofday gmdate gmmktime "
            "gmstrftime localtime microtime mktime strftime strtotime time "
            "dba_close dba_delete dba_exists dba_fetch dba_firstkey filetype "
            "dba_insert dba_list dba_nextkey dba_open dba_optimize dba_popen "
            "dba_replace dba_sync dbase_add_record dbase_close dbase_create "
            "dbase_delete_record dbase_get_record_with_names dbase_get_record "
            "dbase_numfields dbase_numrecords dbase_open dbase_pack filectime "
            "dbase_replace_record dblist dbmclose dbmdelete dbmexists dbmfetch "
            "dbmfirstkey dbminsert dbmnextkey dbmopen dbmreplace basename "
            "chmod chown clearstatcache copy delete dirname disk_free_space "
            "disk_total_space diskfreespace fclose feof fflush fgetc fgetcsv "
            "fgets fgetss file_exists file_get_contents file fileatime ftell "
            "filegroup fileinode filemtime fileowner fileperms filesize popen "
            "flock fnmatch fopen fpassthru fputs fread fscanf fseek fstat stat "
            "ftruncate fwrite glob is_dir is_executable is_file is_link "
            "is_readable is_uploaded_file is_writable is_writeable link "
            "lstat mkdir move_uploaded_file parse_ini_file pathinfo pclose "
            "readfile readlink realpath rename rewind rmdir set_file_buffer "
            "symlink tempnam tmpfile touch umask unlink ftp_cdup ftp_chdir "
            "ftp_close ftp_connect ftp_delete ftp_exec ftp_fget ftp_fput "
            "ftp_get_option ftp_get ftp_login ftp_mdtm ftp_mkdir textdomain "
            "ftp_nb_fget ftp_nb_fput ftp_nb_get ftp_nb_put ftp_nlist ftp_pasv "
            "ftp_put ftp_pwd ftp_quit ftp_rawlist ftp_rename ftp_rmdir checkin "
            "ftp_set_option ftp_site ftp_size ftp_ssl_connect ftp_systype "
            "call_user_func_array call_user_func create_function func_get_arg "
            "func_get_args func_num_args function_exists get_defined_functions "
            "register_shutdown_function register_tick_function method_exists "
            "unregister_tick_function bind_textdomain_codeset bindtextdomain "
            "dcgettext dcngettext dgettext dngettext gettext ngettext "
            "gmp_abs gmp_add gmp_and gmp_clrbit gmp_cmp gmp_com gmp_div_q "
            "gmp_div_qr gmp_div_r gmp_div gmp_divexact gmp_fact gmp_gcd "
            "gmp_hamdist gmp_init gmp_intval gmp_invert gmp_jacobi gmp_gcdext "
            "gmp_mod gmp_mul gmp_neg gmp_or gmp_perfect_square gmp_popcount "
            "gmp_pow gmp_powm gmp_prob_prime gmp_random gmp_scan0 gmp_scan1 "
            "gmp_setbit gmp_sign gmp_sqrt gmp_sqrtrm gmp_strval gmp_sub "
            "header headers_sent setcookie hw_api_attribute hwapi_hgcsp "
            "hw_api_content hw_api_object key langdepvalue value values insert "
            "checkout children mimetype read content copy dbstat dcstat "
            "dstofsrcanchors count reason find ftstat hwstat identify info "
            "insertanchor insertcollection insertdocument link lock move "
            "attreditable count insert remove title value object dstanchors "
            "parents description type remove replace setcommitedversion assign "
            "srcanchors srcsofdst unlock user userlist iconv_get_encoding "
            "iconv_set_encoding iconv ob_iconv_handler exif_imagetype gmp_xor "
            "exif_read_data exif_thumbnail gd_info getimagesize cpdf_stroke "
            "image_type_to_mime_type image2wbmp imagealphablending imagearc "
            "imagechar imagecharup imagecolorallocate imagecolorallocatealpha "
            "imagecolorat imagecolorclosest imagecolorclosestalpha curl_setopt "
            "imagecolorclosesthwb imagecolordeallocate imagecolorexact "
            "imagecolorexactalpha imagecolorresolve imagecolorresolvealpha "
            "imagecolorset imagecolorsforindex imagecolorstotal dba_handlers "
            "imagecolortransparent imagecopy imagecopymerge imagecopymergegray "
            "imagecopyresampled imagecopyresized imagecreate objectbyanchor "
            "imagecreatefromgd2part imagecreatefromgd imagecreatefromgif "
            "imagecreatefromjpeg imagecreatefrompng imagecreatefromstring "
            "imagecreatefromwbmp imagecreatefromxbm imagecreatefromxpm "
            "imagecreatetruecolor imagedashedline imagedestroy imageellipse "
            "imagefill imagefilledarc imagefilledellipse imagefilledpolygon "
            "imagefilledrectangle imagefilltoborder imagefontheight "
            "imageftbbox imagefttext imagegammacorrect imagegd2 imagegd "
            "imageinterlace imagejpeg imageline imageloadfont imagepalettecopy "
            "imagepng imagepolygon imagepsbbox imagepscopyfont imagefontwidth "
            "imagepsextendfont imagepsfreefont imagepsloadfont gmp_legendre "
            "imagepstext imagerectangle imagerotate imagesetbrush imagegif "
            "imagesetstyle imagesetthickness imagesettile imagestring "
            "imagestringup imagesx imagesy imagetruecolortopalette "
            "imagettftext imagetypes imagewbmp iptcembed iptcparse jpeg2wbmp "
            "png2wbmp read_exif_data imap_8bit imap_alerts imap_append "
            "imap_binary imap_body imap_bodystruct imap_check imap_base64 "
            "imap_close imap_createmailbox imap_delete imap_deletemailbox "
            "imap_errors imap_expunge imap_fetch_overview imap_fetchbody "
            "imap_fetchheader imap_fetchstructure imap_get_quota imagettfbbox "
            "imap_getmailboxes imap_getsubscribed imap_header imap_headerinfo "
            "imap_headers imap_last_error imap_list imap_listmailbox "
            "imap_listsubscribed imap_lsub imap_mail_compose imap_mail_copy "
            "imap_mail_move imap_mail imap_mailboxmsginfo imap_listscan "
            "imap_msgno imap_num_msg imap_num_recent imap_open imap_ping "
            "imap_renamemailbox imap_reopen imap_rfc822_parse_adrlist linkinfo "
            "imap_rfc822_parse_headers imap_rfc822_write_address imap_qprint "
            "imap_search imap_set_quota imap_setacl imap_setflag_full "
            "imap_status imap_subscribe imap_thread imap_uid imap_undelete "
            "imap_unsubscribe imap_utf7_decode imap_utf7_encode imap_utf8 "
            "assert_options assert dl extension_loaded get_cfg_var imap_sort "
            "get_defined_constants get_extension_funcs get_included_files "
            "get_loaded_extensions get_magic_quotes_gpc get_current_user "
            "get_required_files getenv getlastmod getmygid getmyinode getmypid "
            "getmyuid getopt getrusage ini_alter ini_get_all ini_get "
            "ini_set php_ini_scanned_files php_logo_guid php_sapi_name "
            "phpcredits phpinfo phpversion putenv set_magic_quotes_runtime "
            "set_time_limit version_compare zend_logo_guid zend_version "
            "ldap_8859_to_t61 ldap_add ldap_bind ldap_close ldap_compare "
            "ldap_connect ldap_count_entries ldap_delete ldap_dn2ufn php_uname "
            "ldap_errno ldap_error ldap_explode_dn ldap_first_attribute "
            "ldap_first_entry ldap_first_reference ldap_free_result "
            "ldap_get_attributes ldap_get_dn ldap_get_entries ldap_get_option "
            "ldap_get_values_len ldap_get_values ldap_list ldap_mod_add "
            "ldap_mod_del ldap_mod_replace ldap_modify ldap_next_attribute "
            "ldap_next_entry ldap_next_reference ldap_parse_reference hypot "
            "ldap_parse_result ldap_read ldap_rename ldap_search ldap_err2str "
            "ldap_set_option ldap_set_rebind_proc ldap_sort ldap_start_tls "
            "ldap_t61_to_8859 ldap_unbind ezmlm_hash mail abs acos acosh asin "
            "asinh atan2 atan atanh base_convert bindec ceil cos cosh decbin "
            "dechex decoct deg2rad exp expm1 floor fmod getrandmax hexdec "
            "is_finite is_infinite is_nan lcg_value log10 log1p log max min "
            "mt_getrandmax mt_rand mt_srand octdec pi pow rad2deg rand round "
            "sin sinh sqrt srand tan tanh mb_convert_case mb_convert_encoding "
            "mb_convert_kana mb_convert_variables mb_decode_mimeheader "
            "mb_decode_numericentity mb_detect_encoding mb_detect_order "
            "mb_encode_mimeheader mb_encode_numericentity mb_ereg_match "
            "mb_ereg_replace mb_ereg_search_getpos mb_ereg_search_getregs "
            "mb_ereg_search_init mb_ereg_search_pos mb_ereg_search_regs "
            "mb_ereg_search_setpos mb_ereg_search mb_ereg mb_eregi_replace "
            "mb_eregi mb_get_info mb_http_input mb_http_output ini_restore "
            "mb_internal_encoding mb_language mb_output_handler mb_parse_str "
            "mb_preferred_mime_name mb_regex_encoding mb_regex_set_options "
            "mb_send_mail mb_split mb_strcut mb_strimwidth mb_strlen mb_strpos "
            "mb_strrpos mb_strtolower mb_strtoupper mb_strwidth imagesetpixel "
            "mb_substitute_character mb_substr_count mb_substr mcrypt_cbc "
            "mcrypt_cfb mcrypt_create_iv mcrypt_decrypt mcrypt_ecb "
            "mcrypt_enc_get_algorithms_name mcrypt_enc_get_block_size "
            "mcrypt_enc_get_iv_size mcrypt_enc_get_key_size ftp_nb_continue "
            "mcrypt_enc_get_modes_name mcrypt_enc_get_supported_key_sizes "
            "mcrypt_enc_is_block_algorithm_mode mcrypt_enc_is_block_algorithm "
            "mcrypt_enc_is_block_mode mcrypt_enc_self_test mcrypt_encrypt "
            "mcrypt_generic_deinit mcrypt_generic_end mcrypt_generic_init "
            "mcrypt_generic mcrypt_get_block_size mcrypt_get_cipher_name "
            "mcrypt_get_iv_size mcrypt_get_key_size mcrypt_list_algorithms "
            "mcrypt_list_modes mcrypt_module_close imap_scanmailbox "
            "mcrypt_module_get_algo_key_size imap_get_quotaroot "
            "mcrypt_module_is_block_algorithm_mode imap_mime_header_decode "
            "mcrypt_module_is_block_mode mcrypt_module_open imagecreatefromgd2 "
            "mcrypt_ofb mdecrypt_generic mhash_count mhash_get_block_size "
            "mhash_get_hash_name mhash_keygen_s2k mhash mime_content_type "
            "connection_aborted connection_status connection_timeout constant "
            "defined get_browser highlight_file highlight_string "
            "ignore_user_abort pack show_source sleep uniqid unpack usleep "
            "msql_affected_rows msql_close msql_connect msql_create_db define "
            "msql_data_seek msql_dbname msql_drop_db msql_dropdb msql_error "
            "msql_fetch_array msql_fetch_field msql_fetch_object msql_createdb "
            "msql_field_seek msql_fieldflags msql_fieldlen msql_fieldname "
            "msql_fieldtable msql_fieldtype msql_free_result msql_freeresult "
            "msql_list_dbs msql_list_fields msql_list_tables msql_listdbs "
            "msql_listfields msql_listtables msql_num_fields msql_num_rows "
            "msql_numfields msql_numrows msql_pconnect msql_query msql_regcase "
            "msql_result msql_select_db msql_selectdb msql_tablename msql "
            "checkdnsrr closelog debugger_off debugger_on gethostbyaddr "
            "dns_check_record dns_get_mx dns_get_record fsockopen "
            "gethostbyname gethostbynamel getmxrr getprotobyname "
            "getservbyname getservbyport ip2long long2ip openlog pfsockopen "
            "socket_get_status socket_set_blocking socket_set_timeout syslog "
            "ocibindbyname ocicancel OCICollAppend ocicollassign "
            "ocicollgetelem ocicollmax ocicollsize ocicolltrim ocicolumnisnull "
            "ocicolumnname ocicolumnprecision ocicolumnscale ocicolumnsize "
            "ocicolumntype ocicolumntyperaw ocicommit ocidefinebyname ocierror "
            "ociexecute ocifetch ocifetchinto ocifetchstatement msql_fetch_row "
            "ocifreecursor OCIFreeDesc ocifreestatement ociinternaldebug "
            "ocilogoff ocilogon ocinewcollection ocinewcursor ocinewdescriptor "
            "ocinlogon ocinumcols ociparse ociplogon ociresult ocirollback "
            "ocirowcount ocisavelob ocisavelobfile ociserverversion ociloadlob "
            "ocisetprefetch ocistatementtype ociwritelobtofile flush ob_clean "
            "ob_end_clean ob_end_flush ob_flush ob_get_contents ob_get_length "
            "ob_get_level ob_get_status ob_gzhandler ob_implicit_flush "
            "overload pcntl_exec pcntl_fork pcntl_signal pcntl_waitpid "
            "pcntl_wexitstatus pcntl_wifexited pcntl_wifsignaled ob_start "
            "pcntl_wstopsig pcntl_wtermsig preg_grep preg_match_all preg_match "
            "preg_quote preg_replace_callback preg_replace preg_split "
            "pdf_add_annotation pdf_add_bookmark pdf_add_launchlink "
            "pdf_add_note pdf_add_outline pdf_add_pdflink pdf_add_thumbnail "
            "pdf_add_weblink pdf_arc pdf_arcn pdf_attach_file pdf_begin_page "
            "pdf_begin_pattern pdf_begin_template pdf_circle pdf_add_locallink "
            "pdf_close_pdi_page pdf_close_pdi pdf_close pcntl_wifstopped "
            "pdf_closepath_stroke pdf_closepath pdf_concat pdf_continue_text "
            "pdf_curveto pdf_delete pdf_end_page pdf_end_pattern "
            "pdf_endpath pdf_fill_stroke pdf_fill pdf_findfont pdf_get_buffer "
            "pdf_get_font pdf_get_fontname pdf_get_fontsize pdf_open_pdi_page "
            "pdf_get_image_width pdf_get_majorversion pdf_get_minorversion "
            "pdf_get_parameter pdf_get_pdi_parameter pdf_get_pdi_value "
            "pdf_initgraphics pdf_lineto pdf_makespotcolor pdf_moveto pdf_new "
            "pdf_open_CCITT pdf_open_file pdf_open_gif pdf_open_image_file "
            "pdf_open_image pdf_open_jpeg pdf_open_memory_image "
            "pdf_open_pdi pdf_open_png pdf_open_tiff pdf_open pdf_place_image "
            "pdf_place_pdi_page pdf_rect pdf_restore pdf_rotate pdf_get_value "
            "pdf_set_border_color pdf_set_border_dash pdf_set_border_style "
            "pdf_set_char_spacing pdf_set_duration pdf_set_font "
            "pdf_set_info_author pdf_set_info_creator pdf_set_info_keywords "
            "pdf_set_info_subject pdf_set_info_title pdf_set_info "
            "pdf_set_parameter pdf_set_text_matrix pdf_set_text_pos "
            "pdf_set_text_rendering pdf_set_text_rise pdf_set_value "
            "pdf_set_word_spacing pdf_setcolor pdf_setdash pdf_setflat "
            "pdf_setgray_fill pdf_setgray_stroke pdf_setgray pdf_setlinecap "
            "pdf_setlinejoin pdf_setlinewidth pdf_setmatrix pdf_setmiterlimit "
            "pdf_setpolydash pdf_setrgbcolor_fill pdf_setrgbcolor_stroke "
            "pdf_setrgbcolor pdf_show_boxed pdf_show_xy pdf_show pdf_skew "
            "pdf_stringwidth pdf_stroke pdf_translate pg_affected_rows "
            "pg_cancel_query pg_client_encoding pg_close pg_connect "
            "pg_connection_busy pg_connection_reset pg_connection_status "
            "pg_copy_from pg_copy_to pg_dbname pg_delete pg_end_copy "
            "pg_escape_string pg_fetch_all pg_fetch_array pg_fetch_assoc "
            "pg_fetch_object pg_fetch_result pg_fetch_row pg_field_is_null "
            "pg_field_name pg_field_num pg_field_prtlen pg_field_size "
            "pg_free_result pg_get_notify pg_get_pid pg_get_result pg_host "
            "pg_last_error pg_last_notice pg_last_oid pg_lo_close pg_lo_create "
            "pg_lo_export pg_lo_import pg_lo_open pg_lo_read_all pg_lo_read "
            "pg_lo_seek pg_lo_tell pg_lo_unlink pg_lo_write pg_meta_data "
            "pg_num_fields pg_num_rows pg_options pg_pconnect pg_ping pg_port "
            "pg_put_line pg_query pg_result_error pg_result_seek pg_field_type "
            "pg_select pg_send_query pg_set_client_encoding pg_trace pg_tty "
            "pg_unescape_bytea pg_untrace pg_update posix_ctermid posix_getcwd "
            "posix_getegid posix_geteuid posix_getgid posix_getgrgid pg_insert "
            "posix_getgroups posix_getlogin posix_getpgid posix_getpgrp "
            "posix_getppid posix_getpwnam posix_getpwuid posix_getrlimit "
            "posix_getuid posix_isatty posix_kill posix_mkfifo posix_setegid "
            "posix_seteuid posix_setgid posix_setpgid posix_setsid pdf_setfont "
            "posix_times posix_ttyname posix_uname pspell_add_to_personal "
            "pspell_add_to_session pspell_check pspell_clear_session "
            "pspell_config_create pspell_config_ignore pspell_config_mode "
            "pspell_config_personal pspell_config_repl posix_setuid "
            "pspell_config_save_repl pspell_new_config pspell_new_personal "
            "pspell_new pspell_save_wordlist pspell_store_replacement "
            "recode_file recode_string recode ereg_replace ereg eregi_replace "
            "split spliti sql_regcase ftok msg_get_queue msg_receive "
            "msg_send msg_set_queue msg_stat_queue sem_acquire sem_get "
            "sem_remove shm_attach shm_detach shm_get_var shm_put_var "
            "shm_remove session_cache_expire session_cache_limiter sem_release "
            "session_destroy session_encode session_get_cookie_params eregi "
            "session_is_registered session_module_name session_name session_id "
            "session_register session_save_path session_set_cookie_params "
            "session_set_save_handler session_start session_unregister "
            "session_write_close snmp_get_quick_print snmp_set_quick_print "
            "snmprealwalk snmpset snmpwalk snmpwalkoid socket_accept snmpget "
            "socket_clear_error socket_close socket_connect session_unset "
            "socket_create_pair socket_create socket_get_option socket_bind "
            "socket_getsockname socket_iovec_add socket_iovec_alloc "
            "socket_iovec_delete socket_iovec_fetch socket_iovec_free "
            "socket_iovec_set socket_last_error socket_listen socket_read "
            "socket_readv socket_recv socket_recvfrom socket_recvmsg "
            "socket_send socket_sendmsg socket_sendto socket_set_nonblock "
            "socket_set_option socket_shutdown socket_strerror socket_write "
            "socket_writev stream_context_create stream_context_get_options "
            "stream_context_set_option stream_context_set_params socket_select "
            "stream_filter_prepend stream_get_filters stream_get_meta_data "
            "stream_get_wrappers stream_register_filter stream_filter_append "
            "stream_select stream_set_blocking stream_set_timeout posix_getpid "
            "stream_set_write_buffer addcslashes addslashes bin2hex chop chr "
            "chunk_split convert_cyr_string count_chars crc32 crypt "
            "fprintf get_html_translation_table hebrev hebrevc explode "
            "htmlentities htmlspecialchars implode join levenshtein localeconv "
            "md5_file md5 metaphone money_format nl_langinfo nl2br pg_convert "
            "parse_str printf quoted_printable_decode quotemeta rtrim "
            "setlocale sha1_file sha1 similar_text soundex sprintf sscanf ord "
            "str_repeat str_replace str_rot13 str_shuffle str_word_count ltrim "
            "strchr strcmp strcoll strcspn strip_tags stripcslashes strcasecmp "
            "stristr strlen strnatcasecmp strnatcmp strncasecmp strncmp strpos "
            "strrchr strrev strrpos strspn strstr strtok strtolower strtoupper "
            "substr_count substr_replace substr trim ucfirst ucwords vprintf "
            "wordwrap base64_decode base64_encode get_meta_tags parse_url "
            "rawurldecode rawurlencode urldecode urlencode doubleval "
            "get_defined_vars get_resource_type gettype stripslashes str_pad "
            "intval is_array is_bool is_callable is_double is_float is_int "
            "is_long is_null is_numeric is_object is_real is_resource floatval "
            "is_string print_r serialize settype strval unserialize "
            "var_dump var_export utf8_decode utf8_encode xml_error_string "
            "xml_get_current_byte_index xml_get_current_column_number "
            "xml_get_current_line_number xml_get_error_code is_scalar vsprintf "
            "xml_parse xml_parser_create_ns xml_parser_create xml_parser_free "
            "xml_parser_get_option xml_parser_set_option number_format "
            "xml_set_default_handler xml_set_element_handler is_integer "
            "xml_set_end_namespace_decl_handler xml_parse_into_struct strtr "
            "xml_set_notation_decl_handler xml_set_object html_entity_decode "
            "xml_set_processing_instruction_handler stream_register_wrapper "
            "xml_set_unparsed_entity_decl_handler xslt_create xslt_errno "
            "xslt_free xslt_output_process xslt_set_base xslt_set_encoding "
            "xslt_set_error_handler xslt_set_log xslt_set_sax_handler "
            "xslt_set_sax_handlers xslt_set_scheme_handler socket_getpeername "
            "zip_close zip_entry_close zip_entry_compressedsize xslt_error "
            "zip_entry_compressionmethod zip_entry_filesize zip_entry_name "
            "zip_entry_open zip_entry_read zip_open zip_read session_decode "
            "get_magic_quotes_runtime xslt_set_scheme_handlers pspell_suggest "
            "xml_set_start_namespace_decl_handler import_request_variables "
            "xml_set_external_entity_ref_handler socket_create_listen "
            "xml_set_character_data_handler session_readonly shm_remove_var "
            "msg_remove_queue pspell_config_runtogether posix_getsid "
            "posix_getgrnam pg_result_status pg_escape_bytea pdf_set_leading "
            "pdf_set_horiz_scaling pdf_save pdf_scale pdf_get_image_height "
            "pdf_end_template pdf_closepath_fill_stroke ocicollassignelem "
            "pdf_clip pdf_close_image ocifreecollection getprotobynumber "
            "mcrypt_module_self_test define_syslog_variables "
            "mcrypt_module_get_supported_key_sizes imap_clearflag_full "
            "mcrypt_module_is_block_algorithm imagepsencodefont "
            "mcrypt_module_get_algo_block_size imagepsslantfont count ")

#---- Syntax Style Specs ----#
SYNTAX_ITEMS = [ (stc.STC_HPHP_DEFAULT,      'default_style'),
                 (stc.STC_HPHP_COMMENT,      'comment_style'),
                 (stc.STC_HPHP_COMMENTLINE,  'comment_style'),
                 (stc.STC_HPHP_COMPLEX_VARIABLE, 'pre_style'), #STYLE ME
                 (stc.STC_HPHP_HSTRING,      'string_style'),
                 (stc.STC_HPHP_HSTRING_VARIABLE, 'scalar_style'), # STYLE ME
                 (stc.STC_HPHP_NUMBER,       'number_style'),
                 (stc.STC_HPHP_OPERATOR,     'operator_style'),
                 (stc.STC_HPHP_SIMPLESTRING, 'string_style'),
                 (stc.STC_HPHP_VARIABLE,     'pre2_style'),
                 (stc.STC_HPHP_WORD,         'keyword_style') ]

#------------------------------------------------------------------------------#

class SyntaxData(syndata.SyntaxDataBase):
    """SyntaxData object for Php""" 
    def __init__(self, langid):
        super(SyntaxData, self).__init__(langid)

        # Setup
        self.SetLexer(stc.STC_LEX_HTML)
        self.RegisterFeature(synglob.FEATURE_AUTOINDENT, AutoIndenter)

    def GetKeywords(self):
        """Returns Specified Keywords List """
        # Support Embedded HTML highlighting
        html = _html.SyntaxData(synglob.ID_LANG_HTML)
        keywords = html.GetKeywords()
        keywords.append((4, PHP_KEYWORDS))
        return keywords

    def GetSyntaxSpec(self):
        """Syntax Specifications """
        return _html.SYNTAX_ITEMS + SYNTAX_ITEMS

    def GetProperties(self):
        """Returns a list of Extra Properties to set """
        return [_html.FOLD, _html.FLD_HTML]

    def GetCommentPattern(self):
        """Returns a list of characters used to comment a block of code
        @note: assuming pure php code for comment character(s)

        """
        return [u'//']

#---- Syntax Modules Internal Functions ----#
def KeywordString(option=0):
    """Returns the specified Keyword String
    @note: not used by most modules

    """
    return PHP_KEYWORDS

#---- End Syntax Modules Internal Functions ----#



###############################################################################
# Name: testHistoryCache.py                                                   #
# Purpose: Unit tests for the base KeyHandler class                           #
# Author: Cody Precord <cprecord@editra.org>                                  #
# Copyright: (c) 2009 Cody Precord <staff@editra.org>                         #
# License: wxWindows License                                                  #
###############################################################################

"""Unittest cases for testing the HistoryCache class"""

__author__ = "Cody Precord <cprecord@editra.org>"
__svnid__ = "$Id: testHistoryCache.py 61142 2009-06-20 05:28:36Z CJP $"
__revision__ = "$Revision: 61142 $"

#-----------------------------------------------------------------------------#
# Imports
import unittest

# Local modules
import common

# Module to test
import ebmlib

#-----------------------------------------------------------------------------#
# Test Class

class HistoryCacheTest(unittest.TestCase):
    def setUp(self):
        self.cache = ebmlib.HistoryCache()

    def tearDown(self):
        pass

    #---- Tests ----#

    def testClear(self):
        """Test clearning the cache"""
        self.cache.PutItem("hello")
        self.assertTrue(self.cache.GetSize() > 0)
        self.cache.Clear()
        self.assertTrue(self.cache.GetSize() == 0)

    def testGetSize(self):
        """Test GetSize method"""
        self.cache.Clear()
        self.assertTrue(self.cache.GetSize() == 0)
        self.cache.PutItem("hello")
        self.assertTrue(self.cache.GetSize() == 1)

    def testGetMaxSize(self):
        """Test retrieving the max size"""
        self.assertTrue(self.cache.GetMaxSize() == ebmlib.HIST_CACHE_UNLIMITED)
        self.cache.SetMaxSize(25)
        self.assertTrue(self.cache.GetMaxSize() == 25)

    def testGetNextItem(self):
        """Test getting the next item in the cache"""
        self.cache.Clear()
        self.cache.PutItem("Hello1")
        self.cache.PutItem("Hello2")
        self.cache.PutItem("Hello3")
        self.cache.PutItem("Hello4")

        # Should be at end of the cache right now
        self.assertTrue(self.cache.GetNextItem() == None, "Found item when none was expected")

        # modify internal pointer for testing purposes
        self.cache.cpos = -1
        item = self.cache.GetNextItem()
        self.assertTrue(item == "Hello1", "Item is %s, pos==%d" % (item, self.cache.cpos))
        item = self.cache.GetNextItem()
        self.assertTrue(item == "Hello2", "Item is %s, pos==%d" % (item, self.cache.cpos))
        item = self.cache.GetNextItem()
        self.assertTrue(item == "Hello3", "Item is %s, pos==%d" % (item, self.cache.cpos))
        item = self.cache.GetNextItem()
        self.assertTrue(item == "Hello4", "Item is %s, pos==%d" % (item, self.cache.cpos))

    def testGetPreviousItem(self):
        """Test getting the previous item in the cache"""
        self.cache.Clear()
        self.cache.PutItem("Hello1")
        self.cache.PutItem("Hello2")
        self.cache.PutItem("Hello3")
        self.cache.PutItem("Hello4")

        self.assertTrue(self.cache.GetPreviousItem() == "Hello4")
        self.assertTrue(self.cache.GetPreviousItem() == "Hello3")
        self.assertTrue(self.cache.GetPreviousItem() == "Hello2")
        self.assertTrue(self.cache.GetPreviousItem() == "Hello1")
        # Should be at beginning of the cache right now
        self.assertTrue(self.cache.GetPreviousItem() == None, "Found item when none was expected")

    def testHasNext(self):
        """Test checking if there are more items in the cache to the right"""
        self.cache.Clear()
        self.assertFalse(self.cache.HasNext(), "HasNext is True in empty cache")
        self.cache.PutItem("Hello1")
        self.cache.GetPreviousItem()
        self.assertTrue(self.cache.HasNext(), "Pos:%d, Size:%d" % (self.cache.cpos, self.cache.GetSize()))

    def testHasPrevious(self):
        """Test checking if there are more items in the cache to the left"""
        self.cache.Clear()
        self.cache.PutItem("Hello1")
        self.assertTrue(self.cache.HasPrevious())
        self.cache.GetPreviousItem()
        self.assertFalse(self.cache.HasPrevious())

    def testPeek(self):
        """Test peeking at the next/prev item in the cache."""
        self.cache.Clear()
        self.assertEquals(self.cache.PeekNext(), None)
        self.assertEquals(self.cache.PeekNext(), None)
        self.assertEquals(self.cache.PeekPrevious(), None)
        self.assertEquals(self.cache.PeekPrevious(), None)
        self.cache.PutItem("hello")
        self.cache.PutItem("hello1")
        self.assertEquals(self.cache.PeekPrevious(), "hello1")
        self.assertEquals(self.cache.PeekNext(), None)
        self.cache.PutItem("hello2")
        self.cache.PutItem("hello3")
        self.assertEquals(self.cache.PeekPrevious(), "hello3")
        self.assertEquals(self.cache.PeekPrevious(), "hello3")
        self.cache.GetPreviousItem() # move back one position
        self.assertEquals(self.cache.PeekPrevious(), "hello2")
        self.assertEquals(self.cache.PeekPrevious(), "hello2")
        self.assertEquals(self.cache.PeekNext(), "hello3")
        self.assertEquals(self.cache.PeekNext(), "hello3")

    def testPutItem(self):
        """Test putting an item in the cache"""
        self.cache.Clear()
        self.cache.PutItem("hello")
        self.assertTrue(self.cache.GetSize() == 1)
        self.cache.PutItem("hello2")
        self.cache.PutItem("hello3")
        self.assertTrue(self.cache.GetSize() == 3)

    def testSetMaxSize(self):
        """Test setting the max size of the cache"""
        self.cache.Clear()

        # Test invalid cache size
        self.assertRaises(AssertionError, self.cache.SetMaxSize, 0)

        # Test setting a valid cache size
        self.cache.SetMaxSize(3)
        self.assertEquals(self.cache.GetMaxSize(), 3)

        # Ensure max size is enforced
        self.cache.PutItem("Hello1")
        self.cache.PutItem("Hello2")
        self.cache.PutItem("Hello3")
        self.assertEquals(self.cache.GetSize(), 3)
        self.cache.PutItem("Hello4")
        self.assertEquals(self.cache.GetSize(), 3)

        # Test that correct items are still on the cache
        self.assertTrue(self.cache.GetPreviousItem(), "Hello4")
        self.assertTrue(self.cache.GetPreviousItem(), "Hello3")
        self.assertTrue(self.cache.GetPreviousItem(), "Hello2")
        item = self.cache.GetPreviousItem()
        self.assertTrue(item == None, "Item should be None: %s" % item)
